# BinomialSampling
#
# Methods and a class to construct binomial samples

# BinomialSampler
#
# A helper type that maintains a cumulative probability list for
# binomial sampling.
type BinomialSampler
  n::Int64;
  p::Float64;
  # n+1 vector of probabilities, prob_vec[i] = P(X == i-1), where X is
  # a Binomial(n, p) random variable.
  prob_vec::Vector{Float64};
  # Vector of cumulative probabilities associated with prob_vec.
  cumulative_prob_vec::Vector{Float64};

  # Construct a new Binomial Sampler with parameters n and p.
  function BinomialSampler(nn::Int64, pp::Float64)
    @assert 0 < nn;
    @assert 0 <= pp <= 1;
    p_vec = zeros(nn + 1);
    ConstructLogProbabilities!(nn, pp, p_vec);
    # Make sure the vector sums to 1 just to be safe.
    p_vec[end] = 1 - sum(p_vec[1:end-1]);
    new(nn, pp, p_vec, cumsum(p_vec));
  end
end  # BinomialSampler

# DrawBinomial(bs::BinomialSampler)
#
# Returns X distributed as Binomial(bs.n, bs.p). Uses an inverse
# transform method, performing binary search to find the point with
# P(X <= i) < U <= P(X < i+1), where X is binomial with parameters
# bs.n and bs.p.
function DrawBinomial(bs::BinomialSampler)
  # bs.cumulative_prob_vec[end] = 1.0;
  high_ind = bs.n + 1;
  low_ind = 1;
  middle_ind = round(Int64, floor((high_ind + low_ind) / 2));
  # Just find the first index of bs.cumulative_prob_vec that satisfies
  # bs.cumulative_prob_vec[index] < U 
  U = rand();
  while (high_ind != low_ind)
    if (U < bs.cumulative_prob_vec[middle_ind])
      # high_ind is too high, so decrease it
      high_ind = middle_ind;
      middle_ind = round(Int64, floor((high_ind + low_ind) / 2));
    else  # U >= bs.cumulative_prob_vec[middle_ind], so middle_ind
          # cannot be the first index. Increase low_ind
      low_ind = middle_ind + 1;
      middle_ind = round(Int64, floor((high_ind + low_ind) / 2));
    end
  end
  return low_ind - 1;
end

# function ConstructLogProbabilities!(n::Int64, p::Float64,
#                                     p_vals::Vector{Float64})
#
# Constructs a probability vector for the Binomial(n, p) distribution,
# which it fills p_vals with. Asserts that p_vals is of length n + 1.
# Upon return, p_vals[i] = P(X == i-1), where X is a Binomial(n, p)
# random variable.
function ConstructLogProbabilities!(nn::Int64, pp::Float64,
                                    p_vals::Vector{Float64})
  @assert length(p_vals) == (nn+1);
  p_vals[:] = 0;
  log_numerator = sum(log(1:nn));
  log_n_m_k_fact = log_numerator;
  log_k_fact = 0;
  log_p = log(pp);
  log_omp = log(1 - pp);
  # Construct log probabilities for everything in the probability
  # vector, then exponential them.
  for ii = 0:nn
    p_vals[ii+1] = (ii * log_p + (nn - ii) * log_omp
                    + log_numerator - log_k_fact - log_n_m_k_fact);
    log_k_fact += log(ii + 1);
    log_n_m_k_fact -= log(nn - ii);
  end
  p_vals[:] = exp(p_vals);
end

# X = ConstructSparseDataset(counts_dict::Dict{String, Int64},
#                            total_count_key::String)
#
# Sets X to be a d-by-n sized sparse 0-1 matrix, where d + 1 is the
# number of keys in counts_dict. Each row of X corresponds to a key
# from counts_dict, and has a number of non-zeros equal to the integer
# in the counts_dict[key] for the corresponding key. These non-zeros
# are permuted randomly. Each column of X is then an observation from
# a process whose counts correspond to counts_dict.
function ConstructSparseDataset(counts_dict::Dict{String, Int64},
                                total_count_key::String)
  key_to_inds = Dict{String, Vector{Int64}}();
  @assert haskey(counts_dict, total_count_key);
  N = counts_dict[total_count_key];
  possible_inds = round(Int64, linspace(1, N, N));
  for key in keys(counts_dict)
    if (key != total_count_key)
      num_for_key = counts_dict[key];
      randperm!(possible_inds, num_for_key);
      key_to_inds[key] = possible_inds[1:num_for_key];
    end
  end
  (row_inds, col_inds) = SparseIndicesFromDict(key_to_inds);
  X = sparse(row_inds, col_inds, 1, key_to_inds.count, N);
  return X;
end

# randperm!(v::Vector, num::Int64)
#
# Sets the first num indices of v to have values chosen from a random
# permutation of v. Runs in time linear in num, and the vector v is
# modified.
function randperm!(v::Vector, num::Int64 = length(v))
  num_left = length(v);
  num = min(num_left, num);
  for ii = 1:num
    swap_ind = ii + round(Int64, floor(num_left * rand()));
    num_left -= 1;
    new_val = v[swap_ind];
    v[swap_ind] = v[ii];
    v[ii] = new_val;
  end
end

# (r, c) = SparseIndicesFromDict(keys_to_indices::Dict{String, Vector{Int64}})
#
# Sets r and c to be vectors of row and column indices--rows for the
# keys in the hash keys_to_indices and columns for the
# values--correspond to the given hash.
function SparseIndicesFromDict(keys_to_indices::Dict{String, Vector{Int64}})
  total_non_zero = 0;
  for key in keys(keys_to_indices)
    total_non_zero += length(keys_to_indices[key]);
  end
  row_inds = Array(Int64, total_non_zero);
  col_inds = Array(Int64, total_non_zero);
  curr_row = 1;
  curr_ind = 1;
  for key in keys(keys_to_indices)
    num_inds_for_key = length(keys_to_indices[key]);
    row_inds[curr_ind:(curr_ind + num_inds_for_key - 1)] = curr_row;
    col_inds[curr_ind:(curr_ind + num_inds_for_key - 1)] =
      keys_to_indices[key][:];
    curr_ind += num_inds_for_key;
    curr_row += 1;
  end
  return (row_inds, col_inds);
end
