# CensorshipExperiment
#
# Author: John Duchi <jduchi@stanford.edu>
#
# Runs an entire censorship experiment from scratch in this script.

include("CensorshipProcessing.jl");
include("LogisticTraining.jl");

module CensorshipExperiments

const CENSOR_LINES = "Data/censored_lines.txt";
const UNCENSOR_LINES = "Data/uncensored_lines.txt";

using LRT;
using Censorship;
using PyPlot;

# Stores the results of a logistic regression experiment
type LogisticExperimentResult
  num_experiments::Int64;
  alpha::Float64;
  prop_train::Float64;
  num_most_frequent::Int64;
  X_all::SparseMatrixCSC;
  y_all::Vector{Float64};
  laplace_losses::Array{Float64,2};
  randomized_response_losses::Array{Float64,2};
  no_privacy_losses::Array{Float64,2};
end  # type LogisticExperimentResult

include("CensorshipSaving.jl");

# alphas = GetPrivacyLevels(num_priv_levels::Int64 = 7)
#
# Differential privacy levels from .1 to something, spaced in powers
# of 2 (lower is more privacy), with num_priv_levels levels.
function GetPrivacyLevels(num_priv_levels::Int64 = 7)
  return logspace(log10(.1), log10(2^(num_priv_levels - 1) * .1),
                  num_priv_levels);
end

function LogregExperiment(num_most_frequent::Int64 = 25;
                          X_all::SparseMatrixCSC = sparse(Array(Int64,0),
                                                          Array(Int64,0),
                                                          Array(Int64,0)),
                          y_all::Vector = Array(Int64,0),
                          num_experiments::Int64 = 20,
                          random_seed::Int64 = 0,
                          alpha = 1.0,
                          prop_train = 0.5,
                          filename = "",
                          cut_dense_at = 15)
  srand(random_seed);
  if (isempty(X_all))
    (y_all, X_all) = MakeFullDataset(CENSOR_LINES, UNCENSOR_LINES);
  end
  n_train = round(Int64, round(prop_train * size(X_all, 1)));
  X_all.nzval[:] = 1;
  X_dense = MakeDenserDataset(X_all, num_most_frequent);
  x_norms = sqrt(sum(X_dense, 2)[:]);
  mults = sqrt(cut_dense_at) ./ x_norms;
  mults[mults .> 1] = 1;
  for ii = 1:size(X_dense, 1)
    X_dense[ii, :] = mults[ii] *  X_dense[ii, :];
  end
  losses_laplace = zeros(2, num_experiments);
  losses_randomized_response = zeros(2, num_experiments);
  losses_no_noise = zeros(2, num_experiments);
  theta_radius = sqrt(num_most_frequent);
  
  for exp_iter = 1:num_experiments
    println("*** Experiment ", exp_iter, " of ", num_experiments, " ***");
    (ytrain, Xtrain, ytest, Xtest) = MakeTrainAndTestSet(y_all, X_dense,
                                                         prop_train);
    println("\tSolving logistic regression with Laplace noise");
    theta_lap = PrivateLogisticSGD(Xtrain, ytrain, theta_radius, alpha,
                                   privacy_style = kLaplacePrivacy);
    println("\tSolving logistic regression with Ltwo noise");
    theta_randomize_response =
      PrivateLogisticSGD(Xtrain, ytrain, theta_radius, alpha,
                         privacy_style = kOptPrivacy);
    println("\tSolving logistic regression with no noise");
    theta_non_private =
      PrivateLogisticSGD(Xtrain, ytrain, theta_radius, Inf,
                         privacy_style = kNoPrivacy);

    l_rr = LogisticObjective(Xtest, ytest, theta_randomize_response);
    l_lap = LogisticObjective(Xtest, ytest, theta_lap);
    l_no_noise = LogisticObjective(Xtest, ytest, theta_non_private);

    losses_randomized_response[1, exp_iter] = l_rr;
    losses_laplace[1, exp_iter] = l_lap;
    losses_no_noise[1, exp_iter] = l_no_noise;

    n_test = size(Xtest, 1);
    t_rr = sum((ytest .* (Xtest * theta_randomize_response)) .<= 0) / n_test;
    t_lap = sum((ytest .* (Xtest * theta_lap)) .<= 0) / n_test;
    t_no_noise = sum((ytest .* (Xtest * theta_non_private)) .<= 0) / n_test;
    losses_randomized_response[2, exp_iter] = t_rr;
    losses_laplace[2, exp_iter] = t_lap;
    losses_no_noise[2, exp_iter] = t_no_noise;
    
    if (!isempty(filename))
      lre = LogisticExperimentResult(num_experiments, alpha,
                                     prop_train,
                                     num_most_frequent,
                                     X_dense, y_all,
                                     losses_laplace,
                                     losses_randomized_response,
                                     losses_no_noise);
      WriteLogisticExperiment(lre, filename);
    end
  end
  lre = LogisticExperimentResult(num_experiments, alpha,
                                 prop_train,
                                 num_most_frequent,
                                 X_dense, y_all,
                                 losses_laplace,
                                 losses_randomized_response,
                                 losses_no_noise);
  if (!isempty(filename))
    WriteLogisticExperiment(lre, filename);
  end
  return lre;
end

# lre = LogisticExperiment(num_most_frequent::Int64 = 800;
#                          X_all::SparseMatrixCSC = sparse(Array(Int64,0),
#                                                         Array(Int64,0),
#                                                         Array(Int64,0)),
#                         y_all::Vector = Array(Int64,0),
#                         num_experiments::Int64 = 20,
#                         random_seed::Int64 = 0,
#                         alpha = .5,
#                         prop_train = 0.1,
#                         filename = "")
#
# Performs a logistic regression experiment, training individual
# predictors for each feature in X.
function LogisticExperiment(num_most_frequent::Int64 = 800;
                            X_all::SparseMatrixCSC = sparse(Array(Int64,0),
                                                            Array(Int64,0),
                                                            Array(Int64,0)),
                            y_all::Vector = Array(Int64,0),
                            num_experiments::Int64 = 20,
                            random_seed::Int64 = 0,
                            alpha = .5,
                            prop_train = 0.1,
                            filename = "")
  srand(random_seed);
  if (isempty(X_all))
    (y_all, X_all) = MakeFullDataset(CENSOR_LINES, UNCENSOR_LINES);
  end
  n_train = round(Int64, round(prop_train * size(X_all, 1)));
  X_all.nzval[:] = 1;
  freq_inds = MostFrequent(X_all, num_most_frequent);
  X_freq = X_all[:, freq_inds];
  losses_laplace = zeros(num_most_frequent, num_experiments);
  losses_randomized_response = zeros(num_most_frequent, num_experiments);
  losses_no_noise = zeros(num_most_frequent, num_experiments);

  for exp_iter = 1:num_experiments
    println("*** Experiment ", exp_iter, " of ", num_experiments, " ***");
    (ytrain, Xtrain, ytest, Xtest) = MakeTrainAndTestSet(y_all, X_freq,
                                                         prop_train);
    println("\tEstimating probabilities with laplace noise");
    counts_lap = LaplaceMultinomialEstimates(Xtrain, ytrain, alpha);
    p_est_lap = zeros(size(counts_lap));
    for ii = 1:size(counts_lap, 2)
      p_est_lap[:, ii] = ProjectOntoSimplex(counts_lap[:, ii], 1.0);
    end
    println("\tEstimating probabilities with Bernoulli noise");
    counts_bern = BernoulliMultinomialEstimates(Xtrain, ytrain, alpha);
    p_est_bern = zeros(size(counts_bern));
    for ii = 1:size(counts_lap, 2)
      p_est_bern[:, ii] = ProjectOntoSimplex(counts_bern[:, ii], 1.0);
    end
    println("\tEstimating probabilities with no noise");
    p_est_no_noise = CountsByCoordinate(Xtrain, ytrain) / size(Xtrain, 1);

    l_rr = LogisticLosses(X_freq, y_all, p_est_bern);
    l_lap = LogisticLosses(X_freq, y_all, p_est_lap);
    l_no_noise = LogisticLosses(X_freq, y_all, p_est_no_noise);

    losses_randomized_response[:, exp_iter] = l_rr;
    losses_laplace[:, exp_iter] = l_lap;
    losses_no_noise[:, exp_iter] = l_no_noise;
    if (!isempty(filename))
      lre = LogisticExperimentResult(num_experiments, alpha,
                                     prop_train,
                                     num_most_frequent,
                                     X_freq, y_all,
                                     losses_laplace,
                                     losses_randomized_response,
                                     losses_no_noise);
      WriteLogisticExperiment(lre, filename);
    end
  end
  lre = LogisticExperimentResult(num_experiments, alpha,
                                 prop_train,
                                 num_most_frequent,
                                 X_freq, y_all,
                                 losses_laplace,
                                 losses_randomized_response,
                                 losses_no_noise);
  if (!isempty(filename))
    WriteLogisticExperiment(lre, filename);
  end
  return lre;
end

# StatisticsOfLogisticExperiment(lre::LogisticExperimentResult;
#                                min_freq::Float64 = 0.0)
#
# Plots a histogram of the ratios of the error ||hat{theta} - theta||
# for the Laplace noise addition technique versus the randomized
# response procedure in the given LogisticExperiment. Only plots
# results for the words with frequency >= min_freq.
#
# Also prints a few statistics of these ratios.
function StatisticsOfLogisticExperiment(lre::LogisticExperimentResult;
                                        min_freq::Float64 = 0.0)
  figure();
  freqs = vec(sum(lre.X_all, 1) / size(lre.X_all, 1));
  inds_to_sample = (freqs .>= min_freq);
  ratios = (lre.laplace_losses[inds_to_sample, :] ./
            lre.randomized_response_losses[inds_to_sample, :]);
  mean_laplace = mean(lre.laplace_losses[inds_to_sample, :], 2);
  mean_r_response = mean(lre.randomized_response_losses[inds_to_sample, :], 2);
  # Construct a histogram with logarithmic spacing to show winners/losers
  figure();
  hist_result = PyPlot.plt[:hist](ratios[:], bins=logspace(-3, 3, 31));
  xscale("log");
  counts = hist_result[1];
  bins = hist_result[2];
  last_le_one_ind = findfirst(bins .>= 1) - 1;
  geo_average_ratio = exp(mean(log(ratios)));
  median_ratio = median(ratios);
  num_laplace_worse = sum(counts[(last_le_one_ind+1):end]);
  num_rr_worse = sum(counts[1:last_le_one_ind]);
  println("*** Statistics ***");
  println("\tGeometric average ratio: ", geo_average_ratio);
  println("\tMedian ratio: ", median_ratio);
  println("\tNumber of times (Laplace, RR) worse: ", 
          round(Int64, num_laplace_worse), ", ", round(Int64, num_rr_worse));
  println("\tNumber of times mean laplace is better than RR: ",
          sum(mean_laplace .< mean_r_response),
          " of ", sum(inds_to_sample));
end

# PlotLogisticExperiment(lre::LogisticExperimentResult;
#                                min_freq::Float64 = 0.0,
#                                basefile::String = "", filetype::String="eps")
#
# Plots the results of the logistic regression experiment stored in
# the data structure lre.
#
# More specifically, plots the mean errors for the Laplace,
# randomized-response, and non-private logistic regression problems
# against the frequency of the appearance of X, the independent
# variable.  The minimum frequency plotted is min_freq. (In the paper,
# min_freq = .015 is used.)
#
# Gives two plots, one of the mean L2 error in the estimate plotted
# against frequency, and the other a plot of mean L2 error in sorted
# order of accuracy.
#
# The basefile is the name of the file saved, filetype is the type of file.
function PlotLogisticExperiment(lre::LogisticExperimentResult;
                                min_freq::Float64 = 0.0,
                                basefile::ASCIIString = "", filetype::ASCIIString="eps")
  figure();
  # Get frequencies of all random variables
  freqs = vec(sum(lre.X_all, 1) / size(lre.X_all, 1));
  mean_lap = mean(lre.laplace_losses, 2)[:];
  mean_rand_resp = mean(lre.randomized_response_losses, 2)[:];
  mean_non_private = mean(lre.no_privacy_losses, 2)[:];
  std_lap = std(lre.laplace_losses, 2)[:] / sqrt(lre.num_experiments);
  std_rand_resp =
    std(lre.randomized_response_losses, 2)[:] / sqrt(lre.num_experiments);;
  std_non_private =
    std(lre.no_privacy_losses, 2)[:] / sqrt(lre.num_experiments);

  inds_to_plot = (freqs .>= min_freq);
  loglog(freqs[inds_to_plot], mean_lap[inds_to_plot], "k-", linewidth=1.5);
  loglog(freqs[inds_to_plot], mean_rand_resp[inds_to_plot],
         "bo-", markersize=5, linewidth=1.5);
  loglog(freqs[inds_to_plot], mean_non_private[inds_to_plot], "rs-");

  ax_coords = [min_freq, maximum(freqs), minimum(mean_non_private) - 1e-3,
               max(maximum(mean_rand_resp[inds_to_plot]),
                   maximum(mean_lap[inds_to_plot])) + .1];
  axis(ax_coords);
  xtick_locs = [ax_coords[1], round(sqrt(ax_coords[1] * ax_coords[2]), 1),
                round(ax_coords[2], 2)];
  ytick_locs = [ax_coords[3],
                (ax_coords[3] * ax_coords[4])^(1.0 / 3),
                (ax_coords[3] * ax_coords[4])^(2.0 / 3),
                ax_coords[4]];
  xticks(xtick_locs, xtick_locs, fontsize=16);
  yticks(fontsize=16);
  legend(("Laplace", "Randomized response", "Non-private"), fontsize=18);
  if (!isempty(basefile))
    if (isempty(filetype))
      savefig(string(basefile, "-freqs.eps"), bbox_inches="tight");
    else
      savefig(string(basefile, "-freqs.", filetype), bbox_inches="tight");
    end
  end

  # Now plot cool standard error plots, but only for sufficiently
  # large frequencies.
  mean_lap = mean_lap[inds_to_plot];
  mean_rand_resp = mean_rand_resp[inds_to_plot];
  mean_non_private = mean_non_private[inds_to_plot];
  std_lap = std_lap[inds_to_plot];
  std_rand_resp = std_rand_resp[inds_to_plot];
  std_non_private = std_non_private[inds_to_plot];
  fill_color = (.85, .85, .8);
  figure();
  len = length(mean_lap);
  inds = unique(len + 1 - round(Int64, polyspace(1, len, 50, 1.5)));
  # sort_inds = sortperm(mean_lap);
  sort_inds = sortperm(mean_rand_resp);
  # sort_inds = sortperm(mean_non_private);
  std_lap[isnan(std_lap)] = 0;
  semilogy((mean_lap[sort_inds])[inds], "kv-", linewidth=2, label="Laplace");
  fill_between(0:(length(inds)-1),
               (mean_lap[sort_inds] + std_lap[sort_inds])[inds],
               (mean_lap[sort_inds] - std_lap[sort_inds])[inds],
               color=fill_color);
  
  # sort_inds = sortperm(mean_rand_resp);
  std_rand_resp[isnan(std_rand_resp)] = 0;
  semilogy((mean_rand_resp[sort_inds])[inds], "bo--", linewidth=2,
           label="Randomized response");
  fill_between(0:(length(inds)-1),
               (mean_rand_resp[sort_inds] + std_rand_resp[sort_inds])[inds],
               (mean_rand_resp[sort_inds] - std_rand_resp[sort_inds])[inds],
               color=fill_color);

  # sort_inds = sortperm(mean_non_private);
  std_non_private[isnan(std_non_private)] = 0;
  semilogy((mean_non_private[sort_inds])[inds], "rs--", linewidth=2,
           label="Non-private");
  fill_between(0:(length(inds)-1),
               (mean_non_private[sort_inds] + std_non_private[sort_inds])[inds],
               (mean_non_private[sort_inds] - std_non_private[sort_inds])[inds],
               color=fill_color);
  legend(fontsize=16);
  xticks(fontsize=16);
  yticks(fontsize=16);
  if (!isempty(basefile))
    if (isempty(filetype))
      savefig(string(basefile, "-sorted.eps"), bbox_inches="tight");
    else
      savefig(string(basefile, "-sorted.", filetype), bbox_inches="tight");
    end
  end
end

# Returns n polynomially spaced values from start to stop.
function polyspace(start::Real, stop::Real, n::Integer,
                   power::Real)
  return linspace(start^(1.0/power), stop^(1.0/power), n).^power;
end

end  # CensorshipExperiments
