# CensorshipProcessing
#
# Author: John Duchi <jduchi@stanford.edu>
#
# Data processing files for reading tokenized documents

module Censorship

import Base.<

export MakeTrainAndTestSet, MakeFullDataset;
export MostFrequent, MakeDenserDataset;

include("IgnoredWords.jl");

const CENS_FILE_CSV = "Data/Censored_Seg.csv";
const UNCENS_FILE_CSV = "Data/NotCensoredSamp_Seg.csv";

# (y_train, X_train, y_test, X_test) =
#    MakeTrainAndTestSet(y_all, X_all;
#                        prop_train::Float64 = .5)
#
# Makes a training set and test set out of the given full label vector
# y and data matrix X, which is of size n-by-d, where d is the ambient
# dimension.
#
# The returned matrix X_train and X_test are of size n_train-by-d and
# n_test-by-d, where d is the ambient data dimension. The split
# proportion into train and test is given by prop_train.
function MakeTrainAndTestSet(y_all, X_all, prop_train::Float64 = .5)
  @assert 0 < prop_train < 1;
  # That is all the data, split it into a training set and a test set.
  (nn, dd) = size(X_all);
  X_all = X_all';
  rand_inds = randperm(nn);
  train_inds = rand_inds[1:round(Int64, floor(prop_train * nn))];
  test_inds = rand_inds[(round(Int64, floor(prop_train * nn)) + 1):end];
  y_train = y_all[train_inds];
  y_test = y_all[test_inds];
  X_train = X_all[:, train_inds];
  X_test = X_all[:, test_inds];
  # Make them all-ones so that it's a 1-hot encoding
  if (typeof(X_all) == SparseMatrixCSC{Float64, Int64})
    X_train.nzval[:] = 1.0;
    X_test.nzval[:] = 1.0;
  end
  return (y_train, X_train', y_test, X_test');
end

# (y, X) = MakeFullDataset(censor_file::String, uncensor_file::String)
#
# Makes a data matrix X and label vector y out of the given files,
# which consist of lines, each of which is a segmented censored or
# uncensored document. The returned X is of size n-by-d, where n is
# the number of examples, d is the dimension of the resulting token
# set.
function MakeFullDataset(censor_file::String, uncensor_file::String,
                         make_zero_one::Bool = false)
  (censored_lines, uncensored_lines) = ReadSegmentedLines(censor_file,
                                                          uncensor_file);
  tokens = Dict{String, Int64}();
  println("\t*** Building token dictionary ***");
  BuildTokenDictionary!(tokens, cat(1, censored_lines, uncensored_lines));
  println("\t*** Building lines to indices ***");
  indices_by_line = LinesToIndices(tokens, cat(1, censored_lines,
                                               uncensored_lines));
  println("\t*** Building sparse data matrices ***");
  (row_inds, col_inds) = LineIndicesToSparseIndices(indices_by_line);
  X_all = sparse(row_inds, col_inds, ones(length(row_inds)));
  y_all = cat(1, ones(length(censored_lines)),
              -ones(length(uncensored_lines)));
  if (make_zero_one)
    X_all.nzval[:] = 1;
  end
  return (y_all, X_all);
end

# indices = MostFrequent(X_all::SparseMatrixCSC,
#                        num_frequent = min(5000, size(X, 1)))
#
# Finds indices of the most frequent words and the indices of the
# num_frequent most frequent coordinates of X, which is assumed to be
# of size n-by-d, where d is the dimension.
function MostFrequent(X_all::SparseMatrixCSC,
                      num_frequent = min(5000, size(X_all, 1)))
  freqs = sum(X_all, 1)[:];
  inds_in_increasing_order = sortperm(freqs);
  return inds_in_increasing_order[end-(num_frequent-1):end];
end

# X_dense = MakeDenserDataset(X::SparseMatrixCSC,
#                             num_frequent = min(100, size(X, 1)))
#
# Finds the num_frequent most frequent features of the sparse {0, 1} matrix X,
# then constructs a dense matrix X_dense consisting of those components of X.
function MakeDenserDataset(X::SparseMatrixCSC,
                           num_frequent = min(100, size(X, 1)))
  inds = MostFrequent(X, num_frequent);
  X_dense = full(X[:, inds]);
  return X_dense;
end

# MakeSegmentedFile(censor_out::String, uncensor_out::String)
#
# Takes as input the destinations for files of censored and uncensored
# documents
function MakeSegmentedFile(censor_out::String,
                           uncensor_out::String)
  cens_data = ReadCSVFile(CENS_FILE_CSV);
  uncens_data = ReadCSVFile(UNCENS_FILE_CSV);
  f_cens = open(censor_out, "w");
  censored_lines = cens_data["textseg"];
  for line in censored_lines
    write(f_cens, line);
    write(f_cens, '\n');
  end
  close(f_cens);
  censored_lines = [];
  uncensored_lines = uncens_data["textseg"];
  f_uncens = open(uncensor_out, "w");
  for line in uncensored_lines
    write(f_uncens, line);
    write(f_uncens, '\n');
  end
  close(f_uncens);
  uncensored_lines = [];
end

# (censored_lines, uncensored_lines) =
#    ReadSegmentedLines(censor_file::String, uncensor_file::String)
#
# Reads the files at [un]censor_file, which consist of lines
# representing blog-posts that have either been censored or not.
# Returns a tuple, which consists of a vector of censored and
# uncensored lines.
function ReadSegmentedLines(censor_file::String, uncensor_file::String)
  f_cens = open(censor_file, "r");
  censored_lines = readlines(f_cens);
  close(f_cens);
  StripEndlineJunk!(censored_lines);
  f_uncens = open(uncensor_file, "r");
  uncensored_lines = readlines(f_uncens);
  close(f_uncens);
  StripEndlineJunk!(uncensored_lines);
  return (censored_lines, uncensored_lines);
end

# StripEndlineJunk!(lines::Vector)
#
# Strips the ending spaces, '\n', and '\r' from the given vector of lines.
function StripEndlineJunk!(lines::Vector)
  for ii = 1:length(lines)
    lines[ii] = lines[ii][1:LastNonEndline(lines[ii],
                                           endline_chars = ['\n', '\r', ' '])];
  end
end

# (row_inds, col_inds) = 
#     LineIndicesToSparseIndices(line_inds::Vector{Vector{Int64}})
#
# Takes a vector mapping from row to vectors of indices and transforms
# it into two vectors row_inds and col_inds, where for any index k, we
# have r = row_inds[k] and c = col_inds[k] if and only if c is
# contained in line_inds[r]. (And all indices in line_inds correspond
# to one such unique index k.)
#
# A sparse matrix S representing line_inds can be constructed as
#
# S = sparse(row_inds, col_inds, ones(length(row_inds)))
function LineIndicesToSparseIndices(line_inds::Vector{Vector{Int64}})
  total_nnz = 0;
  for inds in line_inds
    total_nnz += length(inds);
  end
  curr_count = 1;
  row_inds = zeros(Int64, total_nnz);
  col_inds = zeros(Int64, total_nnz);
  for ii = 1:length(line_inds)
    inds = line_inds[ii];
    for ind in inds
      row_inds[curr_count] = ii;
      col_inds[curr_count] = ind;
      curr_count += 1;
    end
  end
  return (row_inds, col_inds);
end

# indices_by_line = LinesToIndices(tokens::Dict{String, Int64},
#                                  lines::Vector)
#
# Iterates through the given vector of strings, 'lines', adding tokens
# for each to the resulting vector of Vector{Int64}'s, returned as
# indices_by_line.
function LinesToIndices(tokens::Dict{String, Int64},
                        lines::Vector)
  indices_by_line = Array(Vector{Int64}, length(lines));
  for ii = 1:length(lines)
    if (mod(ii, 10000) == 0)
      println("Indexing line ", ii, " of ", length(lines));
    end
    line = lines[ii];
    indices_by_line[ii] = TokenizeLine(tokens, line);
  end
  return indices_by_line;
end

# BuildTokenDictionary!(tokens::Dict{String, Int64},
#                       strings_to_tokenize::Vector)
#
# Takes all the vectors in strings_to_tokenize and adds them to the
# list of tokens that can be hashed.
function BuildTokenDictionary!(tokens::Dict{String, Int64},
                               strings_to_tokenize::Vector)
  for ii = 1:length(strings_to_tokenize)
    if (mod(ii, 10000) == 0)
      println("Tokenizing line ", ii, " of ", length(strings_to_tokenize));
    end
    TokenizeLine!(tokens, strings_to_tokenize[ii]);
  end
end

# indices = TokenizeLine(tokens::Dict{String, Int64},
#                        line::AbstractString;
#                        split_chars::Vector{Char} = [' '])
#
# Computes the indices (tokens) associated with the given string,
# which is split into sub-strings based on split_chars. If an unknown
# token (according to the dictionary 'tokens') is found, throws an
# error. Returns a vector indices of the
# tokens appearing in 'line'.
function TokenizeLine(tokens::Dict{String, Int64},
                      line::AbstractString;
                      split_chars::Vector{Char} = [' '])
  split_line = split(line, split_chars, limit=round(Int64, 1e6), keep=false);
  int_line = Array(Int64, length(split_line));
  for ii = 1:length(split_line)
    ss = split_line[ii];
    wt = GetWordType(ss);
    if (wt == tURL)
      ss = tURLString;
    elseif (wt == tPunctuation)
      ss = tPunctuationString;
    elseif (wt == tEmail)
      ss = tEmailString;
    elseif (wt == tNumber)
      ss = tNumberString;
    end
    # The string is a word enough for us.
    if (haskey(tokens, ss))
      int_line[ii] = tokens[ss];
    else
      error("Dictionary does not have string ", ss);
    end
  end
  return int_line;
end

# indices = TokenizeLine!(tokens::Dict{String, Int64},
#                         line::AbstractString;
#                         split_chars::Vector{Char} = [' '])
#
# Computes the indices (tokens) associated with the given string,
# which is split into sub-strings based on split_chars. Unknown
# tokens, according to the dictionary 'tokens', are added to the
# dictionary with new numbers associated with them. The first token
# has value 1. Returns a vector indices of the tokens appearing in 'line'.
function TokenizeLine!(tokens::Dict{String, Int64},
                       line::AbstractString;
                       split_chars::Vector{Char} = [' '])
  curr_token = tokens.count + 1;
  split_line = split(line, split_chars, limit=round(Int64, 1e6), keep=false);
  int_line = Array(Int64, length(split_line));
  for ii = 1:length(split_line)
    ss = split_line[ii];
    wt = GetWordType(ss);
    if (wt == tURL)
      ss = tURLString;
    elseif (wt == tPunctuation)
      ss = tPunctuationString;
    elseif (wt == tEmail)
      ss = tEmailString;
    elseif (wt == tNumber)
      ss = tNumberString;
    end
    # The string is a word enough for us.
    if (haskey(tokens, ss))
      int_line[ii] = tokens[ss];
    else
      tokens[ss] = curr_token;
      curr_token += 1;
    end
  end
  return int_line;
end

# data = ReadCSVFile(filename::String;
#                    quotemarks::Vector{Char} = ['\"'],
#                    separator::Char = ',',
#                    header::Bool = true)
#
# Reads the comma-separated file stored at the given filename. Returns
# a dictionary of type Dict{String, Vector{String}} that maps the
# header names to vectors of strings, where each vector has 1 entry
# per line. See the method ParseLine for the description of parsing.
function ReadCSVFile(filename::String;
                     quotemarks::Vector{Char} = ['\"'],
                     separator::Char = ',',
                     header::Bool = true)
  data = Dict{String,Vector{String}}();
  f = open(filename, "r");
  mark(f);  # To start at the beginning again after counting lines
  all_lines = readlines(f);
  close(f);
  keys = [];
  if (header)
    first_line = all_lines[1];
    keys = ParseLine(first_line, quotemarks, separator);
  else
    first_line = all_lines[1];
    s = ParseLine(first_line, quotemarks, separator);
    num_keys = length(s);
    keys = map(string, 1:num_keys);
  end
  for ii = 1:length(keys)
    data[keys[ii]] = Array(String,
                           length(all_lines) + (header ? -1 : 0));
  end
  header_ind_fix = (header ? 1 : 0);
  for ii = 1:(length(all_lines) - header_ind_fix)
    parsed_line = ParseLine(all_lines[ii + header_ind_fix],
                            quotemarks, separator);
    if (length(parsed_line) != length(keys))
      error("Parsed line ", all_lines[ii + header_ind_fix],
            " wrong length.\n");
    end
    for kk = 1:length(keys)
      data[keys[kk]][ii] = parsed_line[kk];
    end
  end
  return data;
end

# final_pos = LastNonEndline(line::AbstractString;
#                            endline_chars = ['\n', '\r'])
#
# Returns the last position in the string 'line' that is not one of
# the specified endline characters.
function LastNonEndline(line::AbstractString;
                        endline_chars = ['\n', '\r'])
  final_pos = endof(line);
  while (line[final_pos] in endline_chars && final_pos > 1)
    final_pos = prevind(line, final_pos);
  end
  return final_pos;
end

# s = ParseLine(line::String, quotemarks::Vector{Char},
#               separator::Char;
#               endline_chars = ['\n', '\r', ' '])
#
# Parses the given line and returns a vector of strings 's' containing
# the parsed out words. The parsing functionality first removes all
# end characters from the string, then separates the string into
# constituent words. A word consists of the text between quoted
# characters, where between quoted characters of the same type, all
# special characters--including different quotes and the separator
# character--are ignored. Also, quote characters are ignored if they
# are not followed by a separator character or end of line. So, for
# example, if we parse the string
#
# line = "\"abcd\"abcd\",    , 73,'\"',"\',what\"   "
#
# with quotemarks array of ['\"', '\''] and separator character ','
# we would be returned the list
#
#  "abcd\"abcd"
#  "    "      
#  " 73"       
#  "\""        
#  "',what"
#
# Note in particular that " 73" is treated as a string. The line
# parser also handles unicode encodings.
function ParseLine(line::String,
                   quotemarks::Vector{Char},
                   separator::Char;
                   endline_chars = ['\n', '\r', ' '])
  split_line = String[];
  # Figure out how many end line characters there are and use that as
  # last position.
  final_pos = LastNonEndline(line, endline_chars = endline_chars);
  curr_pos = 1;
  curr_char_list = Char[];
  in_quotes = false;
  curr_quote_mark = char(0);
  while (curr_pos <= final_pos)
    curr_char = line[curr_pos];
    if (curr_char in quotemarks)
      # If the current character is the current quote mark, then we
      # treat this as the end of the quote, as long as one of the
      # following is true: the next character is either (1) the
      # separator or (2) we are at the end of the line.
      if (in_quotes)
        if (curr_quote_mark == curr_char &&
            (curr_pos == final_pos ||
            line[nextind(line, curr_pos)] == separator))
          push!(split_line, convert(String, curr_char_list));
          curr_char_list = Char[];
          # Swap quote status
          in_quotes = false;
          curr_quote_mark = char(0);
        else
          # We are currently in quotes, but the quote mark did not
          # match what started the quoted section, so we treat it as a
          # regular character
          push!(curr_char_list, curr_char);
        end
      else
        # Start the quote action; we weren't quoted already
        in_quotes = true;
        curr_quote_mark = curr_char;
      end
    else
      if (in_quotes)
        # Push character onto the current string
        push!(curr_char_list, curr_char);
      else
        # Not in quotes, but we should be building our character list
        # until we hit the separator
        if (separator == curr_char)
          # This is the separator, finish the word
          if(!isempty(curr_char_list))
            push!(split_line, convert(String, curr_char_list));
            curr_char_list = Char[];
          end
        else
          # Push it onto the current list of characters
          push!(curr_char_list, curr_char);
        end
      end  # if(in_quotes)
    end
    curr_pos = nextind(line, curr_pos);
  end
  if (in_quotes)
    error("Finished parsing line ", line, " while in quotes");
  elseif (!isempty(curr_char_list))
    # If the current character list has some data, push it
    push!(split_line, convert(String, curr_char_list));
  end
  return split_line;
end

end  # module Censorship
