# WriteLogisticExperiment(e_result::LogisticExperimentResult,
#                                 filename::String)
#
# Writes the results of the given logistic regression experiment,
# represented by e_result, to the file stored at the given
# location. Overwrites the stored file.
function WriteLogisticExperiment(e_result::LogisticExperimentResult,
                                 filename::ASCIIString)
  f = open(filename, "w");
  write(f, e_result.num_experiments);
  write(f, e_result.alpha);
  write(f, e_result.prop_train);
  write(f, e_result.num_most_frequent);
  inds = find(e_result.X_all);
  write(f, size(e_result.X_all, 1));
  write(f, size(e_result.X_all, 2));
  write(f, length(inds));
  write(f, inds);
  write(f, length(e_result.y_all));
  write(f, e_result.y_all);
  write(f, size(e_result.laplace_losses, 1));
  write(f, size(e_result.laplace_losses, 2));
  write(f, e_result.laplace_losses);
  write(f, size(e_result.randomized_response_losses, 1));
  write(f, size(e_result.randomized_response_losses, 2));
  write(f, e_result.randomized_response_losses);
  write(f, size(e_result.no_privacy_losses, 1));
  write(f, size(e_result.no_privacy_losses, 2));
  write(f, e_result.no_privacy_losses);
  close(f);
end

# lre = ReadLogisticExperiment(filename::String)
#
# Reads the data stored at the given file, which represents an
# experiment with logistic regression data.
function ReadLogisticExperiment(filename::ASCIIString)
  f = open(filename, "r");
  num_experiments = read(f, Int64);
  alpha = read(f, Float64);
  prop_train = read(f, Float64);
  num_most_frequent = read(f, Int64);
  n = read(f, Int64);
  d = read(f, Int64);
  num_non_zero = read(f, Int64);
  inds = read(f, Int64, num_non_zero);
  (i_inds, j_inds) = ind2sub((n, d), inds);
  X_all = sparse(i_inds, j_inds, ones(length(i_inds)));
  n_y = read(f, Int64);
  @assert n == n_y;
  y_all = read(f, Float64, n_y);
  d1 = read(f, Int64);
  d2 = read(f, Int64);
  laplace_losses = read(f, Float64, (d1, d2));
  d1 = read(f, Int64);
  d2 = read(f, Int64);
  randomized_response_losses = read(f, Float64, (d1, d2));
  d1 = read(f, Int64);
  d2 = read(f, Int64);
  no_privacy_losses = read(f, Float64, (d1, d2));
  return LogisticExperimentResult(num_experiments, alpha, prop_train,
                                  num_most_frequent,
                                  X_all, y_all, laplace_losses,
                                  randomized_response_losses,
                                  no_privacy_losses);
end

# LoadCensorshipExperiment(filename::String)
#
# Loads the data that is the result of a censorship experiment (see
# above) from the given file. Returns the tuple
#
# (train_error, test_error, train_logistic, test_logistic,
#  lap_train_error, lap_test_error, lap_train_logistic, lap_test_logistic,
#  no_train_error, no_test_error, no_train_logistic, no_test_logistic)
function LoadCensorshipExperiment(filename::ASCIIString)
  f = open(filename, "r");
  # Read optimal sampling test results
  d1 = read(f, Int64);
  d2 = read(f, Int64);
  train_error = read(f, Float64, (d1, d2));
  d1 = read(f, Int64);
  d2 = read(f, Int64);
  test_error = read(f, Float64, (d1, d2));
  d1 = read(f, Int64);
  d2 = read(f, Int64);
  train_logistic = read(f, Float64, (d1, d2));
  d1 = read(f, Int64);
  d2 = read(f, Int64);
  test_logistic = read(f, Float64, (d1, d2));
  # Read Laplacian test results
  d1 = read(f, Int64);
  d2 = read(f, Int64);
  lap_train_error = read(f, Float64, (d1, d2));
  d1 = read(f, Int64);
  d2 = read(f, Int64);
  lap_test_error = read(f, Float64, (d1, d2));
  d1 = read(f, Int64);
  d2 = read(f, Int64);
  lap_train_logistic = read(f, Float64, (d1, d2));
  d1 = read(f, Int64);
  d2 = read(f, Int64);
  lap_test_logistic = read(f, Float64, (d1, d2));
  # Read no-privacy test results
  d1 = read(f, Int64);
  no_train_error = read(f, Float64, d1);
  d1 = read(f, Int64);
  no_test_error = read(f, Float64, d1);
  d1 = read(f, Int64);
  no_train_logistic = read(f, Float64, d1);
  d1 = read(f, Int64);
  no_test_logistic = read(f, Float64, d1);
  close(f);
  return (train_error, test_error, train_logistic, test_logistic,
          lap_train_error, lap_test_error,
          lap_train_logistic, lap_test_logistic,
          no_train_error, no_test_error,
          no_train_logistic, no_test_logistic);
end

# SaveCensorshipExperiment
#
# Saves the data from a 
function SaveCensorshipExperiment(train_error::Array{Float64,2},
                                  test_error::Array{Float64,2},
                                  train_logistic::Array{Float64,2},
                                  test_logistic::Array{Float64,2},
                                  lap_train_error::Array{Float64,2},
                                  lap_test_error::Array{Float64,2},
                                  lap_train_logistic::Array{Float64,2},
                                  lap_test_logistic::Array{Float64,2},
                                  no_train_error::Array{Float64,1},
                                  no_test_error::Array{Float64,1},
                                  no_train_logistic::Array{Float64,1},
                                  no_test_logistic::Array{Float64,1},
                                  filename::ASCIIString)
  f = open(filename, "w");
  # Write optimal sampling test results
  write(f, size(train_error, 1));
  write(f, size(train_error, 2));
  write(f, train_error);
  write(f, size(test_error, 1));
  write(f, size(test_error, 2));
  write(f, test_error);
  write(f, size(train_logistic, 1));
  write(f, size(train_logistic, 2));
  write(f, train_logistic);
  write(f, size(test_logistic, 1));
  write(f, size(test_logistic, 2));
  write(f, test_logistic);
  # Write Laplacian test results
  write(f, size(lap_train_error, 1));
  write(f, size(lap_train_error, 2));
  write(f, lap_train_error);
  write(f, size(lap_test_error, 1));
  write(f, size(lap_test_error, 2));
  write(f, lap_test_error);
  write(f, size(lap_train_logistic, 1));
  write(f, size(lap_train_logistic, 2));
  write(f, lap_train_logistic);
  write(f, size(lap_test_logistic, 1));
  write(f, size(lap_test_logistic, 2));
  write(f, lap_test_logistic);
  # Write no-privacy test results
  write(f, size(no_train_error, 1));
  write(f, no_train_error);
  write(f, size(no_test_error, 1));
  write(f, no_test_error);
  write(f, size(no_train_logistic, 1));
  write(f, no_train_logistic);
  write(f, size(no_test_logistic, 1));
  write(f, no_test_logistic);
  close(f);
end
