# GenerateAllExperiments.jl
#
# This single file, if executed in Julia (julialang.org), will
# generate all plots and figures associated with the paper. This file
# may take a while to run. You must have a working version of PyPlot
# installed in your Julia for this to work as well.
#
# Author: jduchi@stanford.edu (John Duchi)

# --------------- DRUG USE EXPERIMENTS --------------- #

# First, perform the drug sampling experiment. This generates the
# image file drug-plots.eps. (Figure 1 in the paper.)

include("drug-experiments.jl");

(laplace_errs, optimal_errs, non_private_errs) =
  PerformDrugExperiment("drug2004.out");
PlotDrugExperiment(laplace_errs, optimal_errs, non_private_errs,
                   filename = "drug-plots.eps");


# ------------- SALARY MEAN EXPERIMENTS --------------- #

# Mean estimation experiments with Salaries, which saves the two
# images salary_mean-1.0.eps and salary_mean-0.5.eps, which correspond
# to two different privacy levels. (Figure 5 in the paper.)

include("SalaryProcessing.jl");
Salary.PrivateMeanExperiment("Data/salary1.txt", 1.0);
Salary.PrivateMeanExperiment("Data/salary1.txt", 0.5);

# Median salary experiment, generating figures salary_median-1.5.eps,
# salary_median-2.0.eps, salary_median-4.0.eps, salary_median-8.0.eps,
# salary_median-16.0.eps. (Figure 6 in the paper.)

Salary.MedianSalaryExperiment("Data/salary1.txt");

# ------------- LOGISTIC REGRESSION AND CENSORSHIP ----------- #

# Logistic regression for censorship experiments. Generates files
# logistic-predictions-freqs.eps and logistic-predictions-sorted.eps,
# which are Figures 7 and 8 in the paper. The method
# StatisticsOfLogisticExperiment also prints some summary statistics.

include("CensorshipExperiment.jl");
lre = CensorshipExperiments.LogisticExperiment(num_experiments = 50,
                                               alpha = 1.0,
                                               prop_train = .25,
                                               filename = "logistic-censor.out");

CensorshipExperiments.StatisticsOfLogisticExperiment(lre, min_freq = .015);
CensorshipExperiments.PlotLogisticExperiment(lre, min_freq = .015,
                                             basefile = "logistic-predictions");

# ----------- LOGISTIC REGRESSION AND CENSORSHIP 2 ---------- #

# True logistic regression using features appearing in at least .5% of
# documents for censorship experiments.

include("CensorshipExperiment.jl");
(y_all, X_all) = Censorship.MakeFullDataset(CensorshipExperiments.CENSOR_LINES,
                                            CensorshipExperiments.UNCENSOR_LINES);

X_all.nzval[:] = 1;
word_counts = sort(full(sum(X_all, 1))[:], lt = >);
prop_documents_appearing = .005;
num_words = findfirst(word_counts .<= prop_documents_appearing * size(X_all, 1));

lre = CensorshipExperiments.LogregExperiment(num_words, alpha = 1.0,
                                             prop_train = .75,
                                             num_experiments = 25,
                                             X_all = X_all,
                                             y_all = y_all,
                                             cut_dense_at = 10,
                                             filename =
                                             "highdim-logreg-censor-alpha1.out");

lre = CensorshipExperiments.LogregExperiment(num_words, alpha = 2.0,
                                             prop_train = .75,
                                             num_experiments = 25,
                                             X_all = X_all,
                                             y_all = y_all,
                                             cut_dense_at = 10,
                                             filename =
                                             "highdim-logreg-censor-alpha2.out");

lre = CensorshipExperiments.LogregExperiment(num_words, alpha = 4.0,
                                             prop_train = .75,
                                             num_experiments = 25,
                                             X_all = X_all,
                                             y_all = y_all,
                                             cut_dense_at = 10,
                                             filename =
                                             "highdim-logreg-censor-alpha4.out");

println("\$\\diffp\$ & Non-private & Optimal & Laplace \\\\");
println("\\hline \\hline");
for ii = [1, 2, 4]
  lre = CensorshipExperiments.ReadLogisticExperiment(
    string("highdim-logreg-censor-alpha", ii, ".out"));
  println(ii, " & \$",
          round(mean(lre.no_privacy_losses[2, :]), 3),
          " \\pm ",
          round(std(lre.no_privacy_losses[2, :]) / sqrt(lre.num_experiments), 3),
          "\$ & \$",
          round(mean(lre.randomized_response_losses[2, :]), 3),
          " \\pm ",
          round(std(lre.randomized_response_losses[2, :])
                / sqrt(lre.num_experiments), 3),
          "\$ & \$",
          round(mean(lre.laplace_losses[2, :]), 3),
          " \\pm ",
          round(std(lre.laplace_losses[2, :]) / sqrt(lre.num_experiments), 3),
          "\$ \\\\");
  println("\\hline");
end

for ii = [1, 2, 4]
  lre = CensorshipExperiments.ReadLogisticExperiment(
    string("highdim-logreg-censor-alpha", ii, ".out"));
  n_exp = lre.num_experiments;
  n_wins = sum(lre.laplace_losses[2, :] .< lre.randomized_response_losses[2,:]);
  p_value = 2 * exp(-2 * n_exp * (n_wins / n_exp - .5)^2);
  println("alpha = ", ii, "; number of Laplace victories: ",
          n_wins, " p = ", p_value);
end

# ----------- LOGISTIC REGRESSION AND CENSORSHIP 3 ---------- #

# True logistic regression using features appearing in at least 10% of
# documents for censorship experiments.

include("CensorshipExperiment.jl");
(y_all, X_all) = Censorship.MakeFullDataset(CensorshipExperiments.CENSOR_LINES,
                                            CensorshipExperiments.UNCENSOR_LINES);

X_all.nzval[:] = 1;
word_counts = sort(full(sum(X_all, 1))[:], lt = >);
prop_documents_appearing = .1;
num_words = findfirst(word_counts .<= prop_documents_appearing * size(X_all, 1));

lre = CensorshipExperiments.LogregExperiment(num_words, alpha = 1.0,
                                             prop_train = .75,
                                             num_experiments = 25,
                                             X_all = X_all,
                                             y_all = y_all,
                                             cut_dense_at = 10,
                                             filename =
                                             "lowdim-logreg-censor-alpha1.out");

lre = CensorshipExperiments.LogregExperiment(num_words, alpha = 2.0,
                                             prop_train = .75,
                                             num_experiments = 25,
                                             X_all = X_all,
                                             y_all = y_all,
                                             cut_dense_at = 10,
                                             filename =
                                             "lowdim-logreg-censor-alpha2.out");

lre = CensorshipExperiments.LogregExperiment(num_words, alpha = 4.0,
                                             prop_train = .75,
                                             num_experiments = 25,
                                             X_all = X_all,
                                             y_all = y_all,
                                             cut_dense_at = 10,
                                             filename =
                                             "lowdim-logreg-censor-alpha4.out");

println("\$\\diffp\$ & Non-private & Optimal & Laplace \\\\");
println("\\hline \\hline");
for ii = [1, 2, 4]
  lre = CensorshipExperiments.ReadLogisticExperiment(
    string("lowdim-logreg-censor-alpha", ii, ".out"));
  println(ii, " & \$",
          round(mean(lre.no_privacy_losses[2, :]), 3),
          " \\pm ",
          round(std(lre.no_privacy_losses[2, :]) / sqrt(lre.num_experiments), 3),
          "\$ & \$",
          round(mean(lre.randomized_response_losses[2, :]), 3),
          " \\pm ",
          round(std(lre.randomized_response_losses[2, :])
                / sqrt(lre.num_experiments), 3),
          "\$ & \$",
          round(mean(lre.laplace_losses[2, :]), 3),
          " \\pm ",
          round(std(lre.laplace_losses[2, :]) / sqrt(lre.num_experiments), 3),
          "\$ \\\\");
  println("\\hline");
end

for ii = [1, 2, 4]
  lre = CensorshipExperiments.ReadLogisticExperiment(
    string("lowdim-logreg-censor-alpha", ii, ".out"));
  n_exp = lre.num_experiments;
  n_wins = sum(lre.laplace_losses[2, :] .< lre.randomized_response_losses[2,:]);
  p_value = 2 * exp(-2 * n_exp * (n_wins / n_exp - .5)^2);
  println("alpha = ", ii, "; number of Laplace victories: ",
          n_wins, " p = ", p_value);
end

