# LogisticTraining
#
# Author: John Duchi <jduchi@stanford.edu>
#
# Methods to train logistic regressions, including random projection
# methods.

module LRT  # LogisticRegressionTraining

include("BinomialSampling.jl");
include("private-vector-sampling.jl");

export kNoPrivacy, kOptPrivacy, kLaplacePrivacy;
export PrivateLogisticSGD, PrivateSVMSGD;
export LogisticObjective, GradientDescentLogreg, ErrorRate;
export BernoulliMultinomialEstimates, LaplaceMultinomialEstimates;
export CountsByCoordinate;
export ChiSquareLosses, LogisticLosses;
export ProjectOntoSimplex;

const kNoPrivacy = "none";
const kOptPrivacy = "opt";
const kLaplacePrivacy = "lap";

# g = LogGrad(X, y, theta::Vector{Float64})
#
# Computes gradient of average logistic objective for n-by-d data
# matrix X for the d-dimensional vector theta. (Data points of X are
# its *rows*).
function LogGrad(X, y, theta::Vector{Float64})
  nn = size(X, 1);
  margins = y .* (X * theta);
  p = 1 ./ (1 + exp(margins));
  g = (1/nn) * (X' * (-y .* p));
  return g[:];
end

# ErrorRate(X, y::Vector, theta::Vector{Float64})
#
# Computes misclassification error rate for the given point theta.
function ErrorRate(X, y::Vector, theta::Vector{Float64})
  nn = size(X, 1);
  return sum(y .!= sign(X * theta)) / nn;
end

# g = LogGrad(x::Vector{Float64}, y::Real, theta::Vector{Float64})
#
# Computes gradient of a single term of the logistic objective for
# d-dimensional data vector x at the point theta.
function LogGrad(x::Vector{Float64}, y::Real, theta::Vector{Float64})
  p = 1 / (1 + exp(y * dot(x, theta)));
  return -y * x * p;
end

# g = SVMGrad(x::Vector{Float64}, y::Real, theta::Vector{Float64})
#
# Computes subgradient of a single term of the hinge-loss objective
# for d-dimensional data vector x at the point theta, where y is the
# label.
function SVMGrad(x::Vector{Float64}, y::Real, theta::Vector{Float64})
  margin = y * dot(x, theta);
  return (margin > 1 ? zeros(size(x)) : -y * x);
end

# mean_theta = PrivateLogisticSGD(X::Array{Float64, 2}, y::Vector,
#                                 radius::Real, alpha::Real;
#                                 optimal_sampling::Bool = true)
#
# Performs private logistic gradient descent with the data given by
# the n-by-d matrix X and n-vector y (labels). Uses dual averaging
# [Nesterov] with optimal stepsize. Returns the average of the
# stochastic gradient iterates, which is the minimax-optimal solution
# for online SGD.
function PrivateLogisticSGD(X::Array{Float64, 2}, y::Vector,
                            radius::Float64, alpha::Real;
                            privacy_style::String = kNoPrivacy)
  (nn, dd) = size(X);
  theta = zeros(dd);
  mean_theta = zeros(dd);
  sum_grads = zeros(dd);
  X_radius = ceil(sqrt(maximum(sum(X.^2, 2))));
  stepsize = (privacy_style == kNoPrivacy ?
              radius / X_radius :
              (privacy_style == kOptPrivacy ?
               radius / PrivateDataRadius(dd, X_radius, alpha) :
               alpha * radius / (dd * X_radius)));
  inds = randperm(nn);
  for ii = 1:nn
    if (mod(ii, 10000) == 0)
      println("SGD iteration ", ii, " of ", nn);
    end
    ind = inds[ii];
    g = LogGrad(X[ind, :][:], y[ind], theta);
    if (privacy_style == kOptPrivacy)
      g = PrivatizeLtwoVector(g, X_radius, alpha);
    elseif (privacy_style == kLaplacePrivacy)
      g = LaplacePrivatizeLtwoVector(g, X_radius, alpha);
    end
    sum_grads += g;
    theta = ProjectOntoLtwoBall(-(stepsize / sqrt(ii)) * sum_grads,
                                radius);
    mean_theta += theta;
  end
  # return theta;
  return mean_theta / nn;
end

function PrivateSVMSGD(X::Array{Float64, 2}, y::Vector,
                       radius::Float64, alpha::Real;
                       privacy_style::String = kNoPrivacy,
                       lambda::Float64 = .05)
  return PrivateSGD(X, y, SVMGrad, radius, alpha, privacy_style, lambda);
end

function PrivateSGD(X::Array{Float64, 2}, y::Vector,
                    grad_func::Function,
                    radius::Float64, alpha::Real,
                    privacy_style::String = kNoPrivacy,
                    lambda::Float64 = .05)
  (nn, dd) = size(X);
  theta = zeros(dd);
  mean_theta = zeros(dd);
  sum_grads = zeros(dd);
  X_radius = ceil(sqrt(maximum(sum(X.^2, 2))));
  stepsize = (privacy_style == kNoPrivacy ?
              radius / X_radius :
              (privacy_style == kOptPrivacy ?
               radius / PrivateDataRadius(dd, X_radius, alpha) :
               alpha * radius / (dd * X_radius)));
  inds = randperm(nn);
  for ii = 1:nn
    if (mod(ii, 10000) == 0)
      println("SGD iteration ", ii, " of ", nn);
    end
    ind = inds[ii];
    g = grad_func(X[ind, :][:], y[ind], theta);
    if (privacy_style == kOptPrivacy)
      g = PrivatizeLtwoVector(g, X_radius, alpha);
    elseif (privacy_style == kLaplacePrivacy)
      g = LaplacePrivatizeLtwoVector(g, X_radius, alpha);
    end
    sum_grads += g;
    if (lambda > 0)
      theta = ProjectOntoLtwoBall(-sum_grads / (lambda * ii), radius);
    else
      theta = ProjectOntoLtwoBall(-(stepsize / sqrt(ii)) * sum_grads,
                                  radius);
    end
    mean_theta += theta;
  end
  # return theta;
  return mean_theta / nn;
end

# LogisticObjective(X, y, theta::Vector{Float64})
#
# Computes and returns average logistic regression objective.  Data
# matrix X is of size n-by-d, where d is the dimension of theta.
function LogisticObjective(X, y, theta::Vector{Float64})
  nn = size(X, 1);
  margins = y .* (X * theta);
  obj = (sum(log(1 + exp(-margins[margins .> 0])))
         + sum(log(1 + exp(margins[margins .<= 0])))
         - sum(margins[margins .<= 0]));
  return obj / nn;
end

# x = ProjectOntoSimplex(v, b)
#
# Sets x to be the minimizer of
#
# norm(x - v, 2)
#
# subject to the constraint that x >= 0 and sum(x) == b. This amounts
# to finding a scalar theta such that sum(max(x - theta, 0)) == b,
# which this function implements (just sort).
function ProjectOntoSimplex(v, b)
  if (b < 0)
    error("Radius of simplex is negative: ", b);
  end
  u = sort(vec(v), lt = >);  # Sort in descending order
  nn = length(u);
  sv = cumsum(u);
  taus = (sv - b) ./ (1 : length(u));
  ind = findfirst(u - taus .< 0);
  ind = (ind == 0 ? nn : ind - 1);
  tau = (sv[ind] - b) / ind;
  return max(v - tau, 0);
end

# ProjectOntoLtwoBall(v::Vector{Float64}, radius::Float64)
#
# Projects the vector v onto the L2 ball of the given radius.
function ProjectOntoLtwoBall(v::Vector{Float64}, radius::Float64)
  return (norm(v, 2) > radius ? radius * v / norm(v, 2) : v)
end

# theta = GradientDescentLogreg(X, y, radius::Float64 = 10.0;
#                               maxiter = 100)
#
# Perform a projected version of gradient descent on the logistic
# regression objective. Uses Armijo (backtracking) line-search with
# projections, and projects to an L2 ball of the given radius.
function GradientDescentLogreg(X, y, radius::Float64 = 10.0;
                               maxiter = 100)
  (nn, dd) = size(X);
  theta = zeros(dd);
  stepsize = 1.0;
  gradmap_size = Inf;
  for iter = 1:maxiter
    println("Iteration ", iter, " of ", maxiter, " (stepsize ",
            round(stepsize, 3), " gradmap size ", round(gradmap_size, 5),
            ")");
    g = LogGrad(X, y, theta);
    curr_obj = LogisticObjective(X, y, theta);
    # Backtracking line search
    expected_decrease = -dot(g[:], g[:]);
    alpha = .2;
    beta = .6;
    while (LogisticObjective(X, y,
                             ProjectOntoLtwoBall(theta - stepsize * g,
                                                 radius)) >
           curr_obj +
           alpha * dot(g, ProjectOntoLtwoBall(theta - stepsize * g, radius)
                       - theta))
      stepsize *= beta;
    end
    new_theta = ProjectOntoLtwoBall(theta - stepsize * g, radius);
    dir = new_theta - theta;
    theta = new_theta;
    gradmap_size = norm(dir) / stepsize;
    if (stepsize < 1e-7)
      break;
    end
    stepsize /= beta;  # Increase the stepsize a bit
  end
  return theta;
end

function ProxL1Step(theta_init::Vector{Float64}, grad::Vector{Float64},
                    stepsize::Float64, lambda::Float64)
  theta = (sign(theta_init - stepsize * grad) .*
           max(abs(theta_init - stepsize * grad) - stepsize * lambda, 0));
  return theta;
end

kMIN_STEPSIZE = 1e-6;

function GradientDescentL1Logreg(X, y, lambda::Float64 = .1;
                                 maxiter = 100)
  (nn, dd) = size(X);
  theta = zeros(dd);
  stepsize = 1;
  old_obj = LogisticObjective(X, y, theta) + lambda * norm(theta, 1);
  for iter = 1:maxiter
    println("Iteration ", iter, " of ", maxiter, " (stepsize ", stepsize,
            ")");
    g = LogGrad(X, y, theta);
    new_theta = ProxL1Step(theta, g, stepsize, lambda);
    new_obj = LogisticObjective(X, y, new_theta) + lambda * norm(new_theta, 1);
    # Backtracking line search
    dir = new_theta - theta;
    alpha = .2;
    beta = .6;
    while (new_obj > old_obj + alpha * dot(g, dir)
           && stepsize > kMIN_STEPSIZE)
      stepsize *= beta;
      new_theta = ProxL1Step(theta, g, stepsize, lambda);
      new_obj = LogisticObjective(X, y, new_theta) + lambda * norm(new_theta, 1);
      dir = new_theta - theta;
    end
    theta = new_theta;
    old_obj = new_obj;
    if (stepsize < kMIN_STEPSIZE)
      break;
    end
    stepsize /= beta;  # Increase the stepsize a bit
  end
  return theta;
end

# counts = CountsByCoordinate(X::SparseMatrixCSC, y::Vector)
#
# Returns a 4-by-d matrix, where X has size n-by-d, of tabulated counts.
# Each column of counts has sum n, and the entries are defined as follows:
#
# counts[1, j] : Number of times coordinate j of X is zero and y is 1
# counts[2, j] : Number of times coordinate j of X is zero and y is -1
# counts[3, j] : Number of times coordinate j of X is non-zero and y is 1
# counts[4, j] : Number of times coordinate j of X is non-zero and y is -1
function CountsByCoordinate(X::SparseMatrixCSC, y::Vector)
  (nn, dd) = size(X);
  # Each of these d-dimensional vectors is a count, by coordinate, of
  # the number of times coordiate x_j was either zero or non-zero,
  # keeping into account the value of y.
  counts_one_one = sum(X[y .== 1, :], 1)[:];
  counts_one_m_one = sum(X[y .== -1, :], 1)[:];
  counts_zero_one = sum(y .== 1) - counts_one_one;
  counts_zero_m_one = sum(y .== -1) - counts_one_m_one;
  counts = [counts_zero_one'; counts_zero_m_one';
            counts_one_one'; counts_one_m_one'];
  return counts;
end

# losses = LogisticLosses(X::SparseMatrixCSC, y::Vector,
#                         estimated_probs::Matrix{Float64},
#                         max_radius::Float64 = log(size(X, 1)))
#
# Takes an n-by-d matrix X, label vector y, and 4-by-d matrix of
# estimated multinomial probabilites (see CountsByCoordinate,
# LaplaceMultinomialEstimates, and
# BernoulliMultinomialEstimates). Returns a vector of length d whose
# j-th entry is the L2 error in the estimated logistic parameters for
# the "true" logistic regression of y against X[:, j] for each of the
# d columns of X.
#
# The logistic parameters are constructed from the 4-by-d matrix of
# estimated probabilities, and each is projected to the l-infinity box
# of radius max_radius.
function LogisticLosses(X::SparseMatrixCSC, y::Vector,
                        estimated_probs::Matrix{Float64},
                        max_radius::Float64 = log(size(X, 1)))
  true_p = CountsByCoordinate(X, y) / size(X, 1);
  losses = zeros(size(true_p, 2));
  theta = zeros(2);
  theta_est = zeros(2);
  for jj = 1:length(losses)
    theta[1] = log(true_p[1, jj] / true_p[2, jj]);
    theta[2] = log(true_p[3, jj] / true_p[4, jj]) - theta[1];
    theta_est[1] = log(estimated_probs[1, jj] / estimated_probs[2, jj]);
    theta_est[2] = (log(estimated_probs[3, jj] / estimated_probs[4, jj])
                    - theta_est[1]);
    inf_inds = isinf(theta_est);
    theta_est[inf_inds] = sign(theta_est[inf_inds]) * max_radius;
    nan_inds = isnan(theta_est);
    theta_est[nan_inds] = 0;
    losses[jj] = norm(theta - theta_est);
  end
  return losses;
end

# losses = ChiSquareLosses(X::SparseMatrixCSC, y::Vector,
#                          estimated_probs::Matrix{Float64})
#
# Takes an n-by-d matrix X, label vector y, and 4-by-d matrix of
# estimated multinomial probabilites (see CountsByCoordinate,
# LaplaceMultinomialEstimates, and
# BernoulliMultinomialEstimates). Returns a vector of length d whose
# jth entry is the chi-squared loss of the estimated probability for
# the true multinomial probability represented in (X, y). The loss for
# an estimate T of parameter theta is
#
# L(T, theta) = (T - theta)^2 / theta.
#
# Each coordinate of losses is the sum of this loss L over coordinate
# j of the true probabilities.
function ChiSquareLosses(X::SparseMatrixCSC, y::Vector,
                         estimated_probs::Matrix{Float64})
  true_p = CountsByCoordinate(X, y) / size(X, 1);
  losses = sum((estimated_probs - true_p).^2 ./ true_p, 1);
  return losses[:];
end

# c = LaplaceMultinomialEstimates(X::SparseMatrixCSC, y::Vector,
#                                 alpha::Float64 = 1.0)
#
# Uses multinomial noise addition to give locally private estimates
# for multinomials. In particular, treats each column of X and
# corresponding vector y as a multinomial over 4 values (paired {0,1}
# values). Adds Laplace noise to (locally-differentially) privatize.
#
# The matrix c returned is of size 4-by-d, where X is of size n-by-d
# and d is the dimension of the problem. Each column of c corresponds to
# a single coordinate of X. So for coordinate j, we have
#
# c[1,j] = Estimated proportion of coordinate j of X is zero, y is 1.
# c[2,j] = Estimated proportion of coordinate j of X is zero, y is -1.
# c[3,j] = Estimated proportion of coordinate j of X is nonzero, y is 1.
# c[4,j] = Estimated proportion of coordinate j of X is nonzero, y is 1.
function LaplaceMultinomialEstimates(X::SparseMatrixCSC, y::Vector,
                                     alpha::Float64 = 1.0)
  (nn, dd) = size(X);
  counts = CountsByCoordinate(X, y);
  # Add the equivalent of Laplace samples to each count. Recall that
  # if X ~ Laplace(a), meaning that it has density (a/2) exp(-a|x|),
  # then X has the same distribution as X_1 - X_2 for X_1, X_2
  # independent Exp(a)-distributed random variables. The sum of n
  # independent Exp(a)-distributed random variables is Gamma(n, a),
  # where a is the rate parameter.
  for jj = 1:dd
    for ind = 1:4
      counts[ind, jj] += RandGamma(nn, alpha) - RandGamma(nn, alpha);
    end
  end
  return counts / nn;
end

# c = BernoulliMultinomialEstimates(X::SparseMatrixCSC, y::Vector,
#                                   alpha::Float64 = 1.0;
#                                   lambda = -1.0)
#
# Uses differentially private randomized response (sign flipping) with
# probabilities p_alpha = e^alpha/(e^alpha+1) to construct an
# estimated frequency matrix F of size 4-by-d, where X is of size
# n-by-d, where each column j of F corresponds to an estimated
# proportion associated with feature j of X. In particular, the matrix
# F satisfies
#
# F[1, j] = Estimated proportion of coordinate j of X is zero, y is 1.
# F[2, j] = Estimated proportion of coordinate j of X is zero, y is -1.
# F[3, j] = Estimated proportion of coordinate j of X is non-zero, y is 1.
# F[4, j] = Estimated proportion of coordinate j of X is non-zero, y is 1.
#
# In expectation, F has the correct mean.
#
# If the value lambda is positive, then hard-thresholds the matrix F
# at the level lambda.
function BernoulliMultinomialEstimates(X::SparseMatrixCSC, y::Vector,
                                       alpha::Float64 = 1.0;
                                       lambda = -1.0)
  (nn, dd) = size(X);
  counts_by_coordinate = zeros(4, dd);
  p_alpha = exp(alpha) / (exp(alpha) + 1);
  B_alpha = (exp(alpha) + 1) / (exp(alpha) - 1);
  for jj = 1:dd
    rand_signs = sign(2 * (rand(nn, 4) .< p_alpha) - 1);
    v = X[:, jj];
    x_one_hot = float([((v .== 0) & (y .== 1))';
                       ((v .== 0) & (y .== -1))';
                       ((v .!= 0) & (y .== 1))';
                       ((v .!= 0) & (y .== -1))']');
    Z = B_alpha * rand_signs .* (x_one_hot - .5) + .5;
    counts_by_coordinate[:, jj] = vec(sum(Z, 1) / nn);
  end
  if (lambda > 0)
    counts_by_coordinate[counts_by_coordinate .<= lambda] = 0;
  end
  return counts_by_coordinate;
end

# X = RandGamma(a::Real, b::Real)
#
# Draws a single Gamma(a, b)-distributed random variable, where a is the
# shape parameter and b is the rate parameter, meaning that X has density
#
# p(x) = (b^a / gamma(a)) * x^(a - 1) exp(-b * x).
#
# Uses rejection sampling.  Assumes that a >= 1. (Smaller b means
# larger variance.)
#
# Reference: A Simple Method for Generating Gamma Variables, Marsaglia
# and Tsang, ACM Transactions on Mathematical Software 26(3),
# pp. 363-372. 2000.
function RandGamma(alpha::Real, beta::Real)
  flag = true;
  d = alpha - 1/3;
  c = 1 / sqrt(9 * d);
  V = 0;
  while (flag)
    Z = randn();
    if (Z >= -1/c)
      U = rand();
      V = (1 + c * Z)^3;
      flag = (log(U) >= .5 * Z^2 + d - d * V + d * log(V));
    end
  end
  return d * V / beta;
end

end  # module LRT

# function BernoulliMultinomialEstimates(X::SparseMatrixCSC, y::Vector,
#                                        alpha::Float64 = 1.0;
#                                        lambda = -1.0)
#   (nn, dd) = size(X);
#   if (lambda < 0)
#     lambda = sqrt(log(nn) / (nn * alpha^2));
#   end
#   counts = CountsByCoordinate(X, y);
#   p_alpha = exp(alpha) / (exp(alpha) + 1);
#   # Number of each of the 4 indices to flip
#   num_flip = sum(rand(nn, 4) .> p_alpha);
#   # A small derivation to show that we need samples from a
#   # hypergeometric distribution.
#   #
#   # Let us assume we have a count n of successes out of N trials.
#   # Suppose that the set F \subset {1, ..., N} contains the indices
#   # that were flipped in the Bernoulli flipping induced by privacy. Then
#   #
#   # Z_i = {  B (1 - 2 X_i)   if i is in F
#   #       {  B (2 X_i - 1)   if i is not in F.
#   #
#   # Then Z_i has mean X_i if B = (e^alpha + 1) / (e^alpha - 1) and the
#   # probability of flipping is 1 / (e^alpha + 1). Moreover, the sum of
#   # the Z_i has the form
#   #
#   # sum_{i in F} B(1 - 2 X_i) + sum_{i in F^c} B(2 X_i - 1)
#   #   = B * (n_F - n_{F^c}) + 2 B X^T (1_{F^c} - 1_F)
#   #
#   # where 1_F denotes the {0, 1} vector with non-zero entries for
#   # indices in F. But as 1_{F^c} - 1_F = 1 - 2 * 1_F, we have that the
#   # preceding display is equal to
#   #
#   #   = B * (n_F - n_{F^c}) + 2 B X^T 1 - 4 B X^T 1_F.
#   #
#   # Noting that if we fix the sum of X and the number of non-zeros in
#   # F, and otherwise the non-zeros of X and F are uniformly
#   # distributed, then X^T 1_F has hypergeometric distribution
#   # conditional on sum(X) and |F|. In particular, the distribution is
#   # Hypergeometric(N, K, n), where K = sum(X) and n = |F| (i.e. there
#   # are K possible successes of N total, and we draw n things.)
#   for ind = 1:4
   
#   end

#   ## Can we cheat and just directly flip the indices in X?
# end
