# MeanEstimation.jl
# Author: John Duchi <jduchi@stanford.edu>
#
# File with methods for estimating means in a differentially private
# way, as well as helper methods for plotting them.

################################################################

# (ests, pows) = EstimatePrivateMeans(X::Array{Float64, 1},
#                                     num_moments::Int64 = 20,
#                                     alpha::Real = 1.0,
#                                     num_tests::Int64 = 10;
#                                     subsample::Float64 = 1.0)
#
# Using the array X, returns private mean estimates using the optimal
# private estimator based on different moment bound assumptions on the
# vector X. In particular, for each of the moment values k in
# logarithmically spaced from {1, ..., log10(3 * num_moments), Inf},
# computes the value
#
# moment = E[|X|^{m_k}]^{1 / m_k},
#
# where m_k is the kth moment value of the logarithmic list above, and
# the expectation is taken with respect to the empirical vector X. In
# particular, we have
#
# m_k = 10^((k - 1) * log10(3 * num_moments) / (num_moments - 1))
#
# as k = 1, 2, ..., num_moments.
#
# Using this moment bound, uses the optimal alpha-locally
# differentially private estimator for the mean under the assumption
# of the given kth moment for k = m_2, ..., m_{num_moments}, Inf to compute
# an estimate of the empirical mean mean(X). This estimate is computed
# using a random 'subsample' fraction of the data.
#
# The above experiment is performed 'num_tests' times, and
# EstimatePrivateMeans returns two quantities:
#
# - A num_tests-by-num_moments matrix ests, which consists of the
#   estimated means
# - A vector pows of length num_moments that corresponds to the powers
#   used in the estimator.
function EstimatePrivateMeans(X::Array{Float64, 1},
                              num_moments::Int64 = 20,
                              alpha::Real = 1.0,
                              num_tests::Int64 = 10;
                              subsample::Float64 = 1.0)
  # For simplicity in setting truncation levels, get true moments for
  # each of the powers m_k above.
  moments = ComputeMoments(X, num_moments);
  mean_estimates = zeros(num_tests, num_moments);
  nn = length(X);
  powers = logspace(0, log10(3 * num_moments), num_moments);
  powers = circshift(powers, -1);
  powers[end] = Inf;
  for iter = 1:num_tests
    println("Performing test ", iter, " of ", num_tests);
    subsampled_x = X;
    if (0 < subsample < 1)
      subsampled_x = X[randperm(nn)[1:round(Int64, ceil(subsample * nn))]];
    end
    for ii = 1:num_moments
      q = powers[ii];
      mean_estimates[iter, ii] = PrivateMean(subsampled_x, q,
                                             moments[ii+1], alpha);
    end
  end
  return (mean_estimates, powers);
end

# mean_estimates = SubsampleMeans(X::Array{Float64, 1},
#                                 num_tests::Int64 = 10;
#                                 subsample::Float64 = 1.0)
#
# Computes num_tests mean estimates of random subsamples of size
# subsample * n of the vector X. Returns all the estimates.
function SubsampleMeans(X::Array{Float64, 1},
                        num_tests::Int64 = 10;
                        subsample::Float64 = 1.0)
  mean_estimates = zeros(num_tests);
  nn = length(X);
  for iter = 1:num_tests
    subsampled_x = X;
    if (0 < subsample < 1)
      subsampled_x = X[randperm(nn)[1:round(Int64, ceil(subsample * nn))]];
    end
    mean_estimates[iter] = mean(subsampled_x);
  end
  return mean_estimates;
end

# Z = DrawLaplaces(N, magnitude)
#
# Draws N i.i.d. Laplace-distributed random variables with E[|Z_i|] =
# magnitude.
function DrawLaplaces(N, magnitude)
  W = rand(N);
  Z = log(W) .* sign(2 * rand(N) - 1);
  return (magnitude * Z);
end

# moments = ComputeMoments(X::Array{Float64, 1}, num_moments::Int64 = 20)
#
# Computes the moments of the vector X treated as an empirical
# distribution. Returns a vector of length num_moments + 1, whose last
# entry is the maximum absolute value in X. The powers are chosen
# logarithmically (see EstimatePrivateMeans for these choices).
function ComputeMoments(X::Array{Float64, 1}, num_moments::Int64 = 20)
  N = length(abs(X));
  moments = zeros(num_moments + 1);
  qvals = logspace(0, log10(3 * num_moments), num_moments);
  # qvals = linspace(1, num_moments, num_moments);
  moments[num_moments + 1] = maximum(abs(X));
  for ii = 1:num_moments
    q = qvals[ii];
    moments[ii] = sum((abs(X) / moments[num_moments+1]).^q) / N;
    moments[ii] = moments[ii]^(1/q) * moments[num_moments+1];
  end
  return moments;
end

# m = PrivateMean(X::Array{Float64, 1}, num_moments,
#                 moment_bound, alpha::Real = 1.0)
#
# Computes estimated mean of data in X privately, using truncation and
# Laplace noise addition.
function PrivateMean(X::Array{Float64, 1}, moment,
                     moment_bound, alpha::Real = 1.0)
  nn = length(X);
  trunc_level = (nn * alpha^2)^(1 / (2 * moment)) * moment_bound;
  trunc_X = max(min(X, trunc_level), -trunc_level);
  W = DrawLaplaces(nn, trunc_level / alpha);
  est_mean = mean(trunc_X + W);
  return est_mean;
end

# PlotMeanExperiment(true_mean::Float64,
#                    estimated_means::Array{Float64, 2},
#                    powers::Array{Float64, 1})
#
# Plots the results of the experiment described in
# PrivateMeanExperiment, using the true value of the mean for the
# data.
function PlotMeanExperiment(true_mean::Float64,
                            estimated_means::Array{Float64, 2},
                            powers::Array{Float64, 1};
                            save_file_name::ASCIIString = "")
  figure();
  num_tests = size(estimated_means, 1);
  num_moments = size(estimated_means, 2);
  mean_error = reshape(mean(abs(estimated_means - true_mean), 1),
                       (num_moments,));
  std_dev = reshape(std(estimated_means - true_mean, 1),
                    (num_moments,));
  # plot_inds = [1:num_moments];
  plot_inds = powers - 1;
  plot_inds[end] = 1.5 * plot_inds[end-1];
  semilogx(plot_inds, mean_error, "b-", linewidth=2);
  fill_between(plot_inds, mean_error - std_dev / sqrt(num_tests),
               mean_error + std_dev / sqrt(num_tests),
               color=[1, .9, .75]);
  semilogx(plot_inds, mean_error, "b-", linewidth=2);
  semilogx(plot_inds, mean_error + std_dev / sqrt(num_tests), "k-.",
       linewidth=1.5);
  semilogx(plot_inds, mean_error - std_dev / sqrt(num_tests), "k-.",
       linewidth=1.5);
  (xmin, xmax, ymin, ymax) = axis();
  min_pow2 = ceil(log2(powers[1] - 1));
  max_pow2 = ceil(log2(powers[end-1]));
  x_tick_inds = 1 + 2.^(min_pow2:max_pow2);
  # xticks(plot_inds[1:2:end], powers[1:2:end]);
  xticks(x_tick_inds-1, x_tick_inds);
  axis([minimum(plot_inds), plot_inds[end], ymin, ymax]);
  if (!isempty(save_file_name))
    savefig(save_file_name, bbox_inches="tight");
  end
end
