# MeanEstimation.jl
# Author: John Duchi <jduchi@stanford.edu>
#
# File with methods for estimating medians in a differentially private
# way, as well as helper methods for plotting them.

################################################################

# ComputeMedianOnline(v::Array{Float64,1}, alpha::Real, max_median::Real)
#
# Uses a stochastic gradient method to compute a locally
# differentially private median of the array v given. Returns a vector
# with the running average of the estimated median computed for each
# iteration of the procedure (the size of the vector is length(v) + 1).
function ComputeMedianOnline(v::Array{Float64,1},
                             alpha::Real,
                             max_median::Real)
  init_guess = max_median * rand();
  p_randomized = exp(alpha) / (exp(alpha) + 1);
  # Flip a count with heads probabilty p_randomized when given a +/- 1
  # gradient.
  B = (exp(alpha) + 1) / (exp(alpha) - 1);
  stepsize_mult = max_median / (2 * B);
  nn = length(v);
  inds = randperm(length(v));
  curr_median = init_guess;
  avg_median = zeros(nn + 1);
  avg_median[1] = curr_median;
  bits = (rand(nn) .< p_randomized);
  for ii = 1:nn
    index = inds[ii];
    # Gradient is sign of curr_median - v[index]
    s = sign(curr_median - v[index]);
    # Perturbed gradient has mean s, but is much larger
    g = B * s * (2 * bits[ii] - 1);
    curr_median = curr_median - g * stepsize_mult / sqrt(ii);
    curr_median = min(max(curr_median, 0), max_median);
    avg_median[ii + 1] = ((ii - 1) * avg_median[ii] + curr_median) / ii;
  end
  return avg_median;
end

# ind = find_first_gt(v::Array{Float64, 1}, x::Real)
#
# Assumes the array v is sorted in increasing order v[1] <= v[2]....
# Finds the first index i such that v[i] >= x, and if no indices
# satisfy this, returns length(v) + 1.
function find_first_gt(v::Array{Float64, 1}, x::Real)
  low_ind = 1;
  high_ind = length(v);
  mid_ind = round(Int64, floor((high_ind + low_ind) / 2));
  while (low_ind < high_ind)
    if (v[mid_ind] >= x)
      # Middle index is high enough, so decrease high index.
      high_ind = mid_ind;
    else  # v[mid_ind] < x, so increase low ind by enough.
      low_ind = mid_ind + 1;
    end
    mid_ind = round(Int64, floor((high_ind + low_ind) / 2));
  end
  if (v[low_ind] < x)
    return length(v) + 1;
  end
  return low_ind;
end

# gaps = MedianLossGaps(estimated_medians::Array{Float64, 1},
#                       v::Array{Float64, 1})
#
# Takes a vector v of samples and a vector 'estimated_medians' of
# estimated medians of v. Computes the gaps between
# loss(estimated_medians[ii]), where
#
# loss(x) = sum(abs(x - v)) / length(v)
#
# and loss(median(v)), the true average absolute distance between a
# point v[i] and the median of v. Returns these gaps.
#
# If m = length(estimated_medians) and n = length(v), run-time of the
# method is O(m * log(n) + n * log(n)). Sorts the vector v in
# increasing order.
function MedianLossGaps(estimated_medians::Array{Float64, 1},
                        v::Array{Float64, 1})
  sort!(v);  # Puts v in increasing order v[1] <= v[2] <= ...
  true_med = median(v);
  sample_size = length(v);
  num_meds = length(estimated_medians);
  min_loss = sum(abs(v - true_med)) / sample_size;
  cumulative_sum_v = cumsum(v);
  cumulative_sum_reverse = cumsum(reverse(v));
  loss_gaps = zeros(size(estimated_medians));
  # Now, compute the losses for each by finding appropriate indices
  # and not actually computing sums.
  for ii = 1:num_meds
    ind = find_first_gt(v, estimated_medians[ii]);
    loss = 0;
    x = estimated_medians[ii];
    if (ind > 1 && ind <= sample_size)
      loss = ((ind - 1) * x - cumulative_sum_v[ind - 1]
              + cumulative_sum_reverse[sample_size - ind + 1]
              - (sample_size - ind + 1) * x) / sample_size;
    elseif (ind == 1)
      loss = (cumulative_sum_v[end] - sample_size * x) / sample_size;
    else  ## ind == sample_size + 1, so all v are < x.
      loss = x - cumulative_sum_v[end] / sample_size;
    end
    loss_gaps[ii] = loss - min_loss;
  end
  return loss_gaps;
end


# (mean_gaps, sub_gaps, var_gaps) =
#           MedianGapsExperiment(X::Array{Float64, 1},
#                                num_tests::Int64 = 100,
#                                alpha = 1;
#                                subsamp_every::Int64 = 100,
#                                truncation_factor::Real = 6.0)
#
# Takes vector of non-negative floats X. For each of num_tests
# iterations, performs an online stochastic gradient descent estimator
# to find the median of X using locally differentially private
# gradient scheme with differential privacy parameter alpha.
#
# Define the (empirical) risk of a median estimator theta to be
#
# R(theta) = sum(abs(X - theta)) / N,
#
# where N is the length of the vector X. For each
# iteration/experiment, this method randomly orders X, then
# computes the gap R(theta) - R(theta*), where theta* is the true
# median. Returns the mean of these gaps as well as the variance
# across all experiments, where entry i of the mean_gaps vector
# corresponds to the i-th iterate of the private SGD procedure.  Also
# returns a subsampled version of the gaps for all experiments in
# 'sub_gaps', subsampled every subsamp_every iterations.
#
# If truncation_factor is finite, sets the maximum allowed value for
# the median as truncation_factor * median(abs(X)). Otherwise the
# maximum value is the maximum value of X.
function MedianGapsExperiment(X::Array{Float64, 1},
                              num_tests::Int64 = 100,
                              alpha = 1;
                              subsamp_every::Int64 = 100,
                              truncation_factor::Real = 6.0)
  if (any(X .< 0))
    error("Assumes positive vector X");
  end
  max_median = maximum(abs(X));
  if (truncation_factor < Inf)
    max_median = truncation_factor * median(abs(X));
  end
  var_gaps = zeros(length(X) + 1);
  mean_loss_gaps = zeros(length(X) + 1);
  subsamp_every = 100;
  subsampled_all_loss_gaps = zeros(round(Int64, floor(length(X)
                                             / subsamp_every)),
                                   num_tests);
  for iter = 1:num_tests
    if (mod(iter, 10) == 0)
      println("Test ", iter, " of ", num_tests);
    end
    meds = ComputeMedianOnline(X, alpha, max_median);
    gaps = MedianLossGaps(meds, X);
    subsampled_all_loss_gaps[:, iter] = gaps[subsamp_every:subsamp_every:end];
    mean_loss_gaps += gaps;
    var_gaps += gaps.^2;
  end
  mean_loss_gaps /= num_tests;
  var_gaps /= num_tests;
  var_gaps -= mean_loss_gaps.^2;
  return (mean_loss_gaps, subsampled_all_loss_gaps, var_gaps);
end

# est = NaiveMedian(X::Array{Float64, 1}, alpha = 1;
#                   allowed_truncation::Real = 6.0)
#
# Using the fact that for a symmetric random variable W and any random
# variable X, the median of X + W is the same as X, returns a naive
# estimator of the mean using locally differentially private
# sampling. In particular, let
#
# tX = max(min(X, T), -T)
#
# denote a truncated version of X, where the truncation level T is set
# as the minimum of the largest X and allowed_truncation * median(abs(X)).
#
# W is chosen to be Laplace(T/alpha) distributed, which guarantees
# alpha-differential privacy.
function NaiveMedian(X::Array{Float64, 1},
                     alpha = 1;
                     allowed_truncation::Real = 6.0)
  range = maximum(abs(X));
  if (allowed_truncation < Inf)
    # If we allow some truncation, then bring down the range a bit.
    range = min(allowed_truncation * median(abs(X)), range);
  end
  nn = length(X);
  Z = (max(min(X, range), -range)
       + (range / alpha) * log(rand(nn)) .* sign(2 * rand(nn) - 1));
  return median(Z);
end

# PlotLossGaps(mean_gaps::Array{Float64, 1},
#              subsampled_all_gaps::Array{Float64, 2},
#              subsamp_size::Int64 = 100;
#              p_val::Float64 = .05,
#              save_file_name = "")
#
# Given the mean of gaps in loss for stochastic gradient iterations
# for a particular procedure and a subsample of all the loss gaps for
# the procedure, plots confidence intervals and the mean on a
# logarithmic scale. Saves the resulting file.
function PlotLossGaps(mean_gaps::Array{Float64, 1},
                      subsampled_all_gaps::Array{Float64, 2},
                      subsamp_size::Int64 = 100;
                      p_val::Float64 = .05,
                      save_file_name = "",
                      additional_gaps::Array{Float64,1} = [])
  upper_loss_gaps = zeros(size(subsampled_all_gaps, 1));
  lower_loss_gaps = zeros(size(subsampled_all_gaps, 1));
  if (p_val > 1 || p_val < 0)
    error("Wrong setting of p-value ", p_val);
  end
  for ii = 1:length(upper_loss_gaps)
    upper_loss_gaps[ii] = quantile(subsampled_all_gaps[ii, :][:], 1 - p_val);
    lower_loss_gaps[ii] = quantile(subsampled_all_gaps[ii, :][:], p_val);
  end
  figure();
  inds_confidences = subsamp_size * (1:length(upper_loss_gaps));
  semilogy(mean_gaps, "b-", linewidth=2);
  fill_between(inds_confidences,
               lower_loss_gaps, upper_loss_gaps,
               color = [1, .9, .75]);
  semilogy(mean_gaps, "b-", linewidth=2);
  semilogy(inds_confidences, lower_loss_gaps, "k-.", linewidth=1.5);
  semilogy(inds_confidences, upper_loss_gaps, "k-.", linewidth=1.5);
  if (!isempty(additional_gaps))
    N = length(mean_gaps);
    mean_gap = mean(additional_gaps);
    upper_gap = quantile(additional_gaps, 1 - p_val);
    lower_gap = quantile(additional_gaps, p_val);
    fill_between([1, N], [lower_gap, lower_gap],
                 [upper_gap, upper_gap], color = [.9, 1, .9]);
    semilogy([1, N], [mean_gap, mean_gap], "-", linewidth = 2,
             color = [0, .2, 0]);
    semilogy([1, N], [upper_gap, upper_gap], "k-.", linewidth=1.5);
    semilogy([1, N], [lower_gap, lower_gap], "k-.", linewidth=1.5);
  end
  (xmin, xmax, ymin, ymax) = axis();
  axis([0, length(mean_gaps), ymin, maximum(mean_gaps)]);
  if (!isempty(save_file_name))
    savefig(save_file_name, bbox_inches="tight");
  end
end
