# SalaryProcessing
#
# Author: John Duchi <jduchi@stanford.edu>
#
# Files for generating experimental results associated with salary
# data. This includes reading a salary file, doing an online SGD with
# optimal privacy-sampling mechanism, and computing statistics of the
# resulting procedure.

module Salary

include("MeanEstimation.jl");
include("MedianEstimation.jl");

# For plotting fun.
using PyPlot;

const kNumTests = 400;
const kNumMoments = 20;

# (ests, pows) = PrivateMeanExperiment(salary_file::ASCIIString, alpha = 1.0)
#
# Using the file provided by 'salary_file', performs an experiment of
# estimating private means, where the different estimators use a
# variety of assumptions on the moments. See EstimatePrivateMeans for
# an explanation of the parameters.
#
# Executing this method with the file "salary1.txt" will generate the
# mean experimental plots in the paper.
function PrivateMeanExperiment(salary_file::ASCIIString, alpha = 1.0)
  srand(0);
  num_moments = kNumMoments;
  num_tests = kNumTests;
  subsample_frac = .5;
  salaries = ReadSalaryFile(salary_file);
  (ests, pows) = EstimatePrivateMeans(salaries, num_moments, alpha,
                                      num_tests, subsample=subsample_frac);
  # Plot mean along with standard errors
  PlotMeanExperiment(mean(salaries), ests, pows,
                     save_file_name = string("salary_mean-", alpha, ".eps"));
  return (ests, pows);
end

# (mean_gaps, sub_gaps, var_gaps, naive_gaps) =
#     MedianSalaryExperiment(salary_file_name::ASCIIString)
#
# Performs an experiment to estimate median salaries, plotting the
# result with 80% confidence intervals (10 on each side). Does this
# experiment for each of the truncation levels 1.5, 2, 4, 8, 16. See
# the method MedianGapsExperiment for more on the parameters. Returns
# the last result of the all the truncation levels.
function MedianSalaryExperiment(salary_file_name::ASCIIString)
  srand(2);
  salaries = ReadSalaryFile(salary_file_name);
  num_tests = 400;
  alpha = 1.0;
  subsamp_every = 100;
  truncation_factors = [1.5, 2, 4, 8, 16];
  mean_gaps = [];
  sub_gaps = [];
  var_gaps = [];
  naive_gaps = [];
  for trunc_ind = 1:length(truncation_factors)
    trunc_factor = truncation_factors[trunc_ind];
    println("\t*** Truncation ", trunc_factor,
            " (", trunc_ind, " of ", length(truncation_factors), ") ***\n");
    naive_medians = zeros(num_tests);
    for ii = 1:num_tests
      if (mod(ii, 10) == 0)
        println("Naive median experiment ", ii, " of ", num_tests);
      end
      naive_medians[ii] = NaiveMedian(salaries, alpha,
                                      allowed_truncation = trunc_factor);
    end
    naive_gaps = MedianLossGaps(naive_medians, salaries);
    (mean_gaps, sub_gaps, var_gaps) =
      MedianGapsExperiment(salaries, num_tests, alpha,
                           subsamp_every = subsamp_every,
                           truncation_factor = trunc_factor);
    save_name = string("salary_median-", trunc_factor, ".eps");
    PlotLossGaps(mean_gaps, sub_gaps, subsamp_every,
                 p_val = .1, save_file_name = save_name,
                 additional_gaps = naive_gaps);
  end
  return (mean_gaps, sub_gaps, var_gaps, naive_gaps);
end

# ReadSalaryFile(fname::ASCIIString)
#
# Reads a list of salaries (one salary per line) from a file name,
# returning them as a float64 vector.
function ReadSalaryFile(fname::ASCIIString)
  f = open(fname);
  lines = readlines(f);
  close(f);
  # Now to parse each line
  for ii = 1:length(lines)
    lines[ii] = strip(lines[ii], '\n');
  end
  salaries = zeros(length(lines));
  for ii = 1:length(lines)
    salaries[ii] = float(lines[ii]);
  end
  return salaries;
end

# SaveMatrix(A::Array{Float64, 2}, filename::ASCIIString)
#
# Helper file for saving a matrix A in binary format. Silently
# overwrites whatever is stored at filename.
function SaveMatrix(A::Array{Float64, 2}, filename::ASCIIString)
  f = open(filename, "w");
  write(f, size(A, 1));
  write(f, size(A, 2));
  write(f, A);
  close(f);
end

# A = ReadMatrix(filename::ASCIIString)
#
# Helper file for reading a matrix in binary format. Returns resulting
# matrix.
function ReadMatrix(filename::ASCIIString)
  f = open(filename, "r");
  d1 = read(f, Int64);
  d2 = read(f, Int64);
  A = read(f, Float64, (d1, d2));
  close(f);
  return A;
end

end  # module Salary
