# drug-experiments.jl
#
# Author: John Duchi <jduchi@stanford.edu>

include("drug-use-data.jl");
include("LogisticTraining.jl");
using PyPlot;

const kNumTests = 100;
const kAlpha = .5;

# (laplace_errs, optimal_errs, non_private_errs)
#     = PerformDrugExperiment(savefile::ASCIIString = "")
#
# Takes all count data from the 2004 drug use study (see the file
# drug-use-data.jl) and simulates sampling from the drug use data.  In
# particular, for each of num_tests = 100 tests, performs the
# following experiment:
#
# 1. Generates a dataset using ConstructSparseDataset (see
# BinomialSampling.jl)
#
# 2. Iterates through 2/3 of the data in the dataset, performing iid
# draws of the observations at each iteration.
#  (2a) Adds either Laplace noise or optimal box-based noise to each
#     vector using PrivatizeLinfVectorLaplace or
#     PrivatizeLinfVector (see the file private-vector-sampling.jl)
#  (2b) Tracks the cumulative mean of sampled data, comparing to the true
#     (population sample) mean
#
# The three matrices returned are of size num_tests-by-(2 N / 3),
# where N is the population size, and correspond to the errors of each
# of the sampling schemes over iterations.
#
# If savefile is specified, saves the results of the experiment in a
# file that can be read by LoadDrugExperiment.
function PerformDrugExperiment(savefile::ASCIIString = "")
  nn = drugs2004["EDvisits"];
  dd = drugs2004.count - 1;
  n_to_sample = round(Int64, (ceil(2 * nn / 3)));
  alpha = kAlpha;
  num_tests = kNumTests;
  X = LRT.ConstructSparseDataset(drugs2004, "EDvisits");
  X = full(float(X));
  (dd, nn) = size(X);

  laplace_max_errs = zeros(num_tests, n_to_sample);
  optimal_max_errs = zeros(num_tests, n_to_sample);
  non_private_max_errs = zeros(num_tests, n_to_sample);
  true_mean = mean(X, 2);

  for test_iter = 1:num_tests
    println("\t*** Test ", test_iter, " of ", num_tests, " ***");
    laplace_means = zeros(dd, n_to_sample);
    optimal_means = zeros(dd, n_to_sample);
    non_private_means = zeros(dd, n_to_sample);
    
    for ii = 1:n_to_sample
      if (mod(ii, 50000) == 0)
        println("Sample ", ii, " of ", n_to_sample);
      end
      laplace_means[:, ii] += LRT.PrivatizeLinfVectorLaplace(X[:, ii], 1.0,
                                                             alpha);
      optimal_means[:, ii] += LRT.PrivatizeLinfVector(X[:, ii], 1.0, alpha);
      non_private_means[:, ii] += X[:, ii];
    end

    laplace_means = cumsum(laplace_means, 2);
    optimal_means = cumsum(optimal_means, 2);
    non_private_means = cumsum(non_private_means, 2);

    for ii = 1:n_to_sample
      laplace_means[:, ii] /= ii;
      optimal_means[:, ii] /= ii;
      non_private_means[:, ii] /= ii;
      laplace_max_errs[test_iter, ii] = norm(laplace_means[:, ii] - true_mean,
                                             Inf);
      optimal_max_errs[test_iter, ii] = norm(optimal_means[:, ii] - true_mean,
                                             Inf);
      non_private_max_errs[test_iter, ii] =
        norm(non_private_means[:, ii] - true_mean, Inf);
    end
  end
  if (!isempty(savefile))
    SaveDrugExperiment(alpha, dd, nn, laplace_max_errs, optimal_max_errs,
                       non_private_max_errs, savefile);
  end
  return (laplace_max_errs, optimal_max_errs, non_private_max_errs);
end

# (alpha, n, d, lap_errs, opt_errs, np_errs) = 
#    LoadDrugExperiment(filename::ASCIIString)
#
# Loads data from a drug use simulation experiment generated by
# PerformDrugExperiment. The parameter alpha is the privacy parameter,
# n is the total population size, d the number of drugs (dimension),
# and lap_errs, opt_errs, and np_errs are num_tests-by-sample_size
# matrices of errors in estimation from the true population mean.
function LoadDrugExperiment(filename::ASCIIString)
  file = open(filename, "r");
  alpha = read(file, Float64);
  num_tests = read(file, Int64);
  n_to_sample = read(file, Int64);
  nn = read(file, Int64);
  dd = read(file, Int64);
  lap_errs = read(file, Float64, (num_tests, n_to_sample));
  opt_errs = read(file, Float64, (num_tests, n_to_sample));
  np_errs = read(file, Float64, (num_tests, n_to_sample));
  close(file);
  return (alpha, nn, dd, lap_errs, opt_errs, np_errs);
end

# SaveDrugExperiment(alpha::Float64, dim::Int64,
#                    N::Int64, laplace_errs::Matrix{Float64},
#                    optimal_errs::Matrix{Float64},
#                    non_private_errs::Matrix{Float64},
#                    filename::ASCIIString)
#
# Saves data from drug experiment
function SaveDrugExperiment(alpha::Float64, dim::Int64,
                            N::Int64, laplace_errs::Matrix{Float64},
                            optimal_errs::Matrix{Float64},
                            non_private_errs::Matrix{Float64},
                            filename::ASCIIString)
  file = open(filename, "w");
  write(file, alpha);
  write(file, size(laplace_errs, 1));
  write(file, size(laplace_errs, 2));
  write(file, N);
  write(file, dim);
  write(file, laplace_errs);
  write(file, optimal_errs);
  write(file, non_private_errs);
  close(file);
end

# PlotDrugExperiment(laplace_errs::Matrix{Float64},
#                    optimal_errs::Matrix{Float64},
#                    non_private_errs::Matrix{Float64};
#                    p_value::Float64 = .1,
#                    filename = "")
#
# Plots results from the drug experiments, with lower and upper
# confidence values of p_value and 1-p_value. Stores the resulting
# plot in filename if specified.
function PlotDrugExperiment(laplace_errs::Matrix{Float64},
                            optimal_errs::Matrix{Float64},
                            non_private_errs::Matrix{Float64};
                            p_value::Float64 = .1,
                            filename = "")
  (num_tests, n_obs) = size(laplace_errs);
  mean_lap_errs = mean(laplace_errs, 1)[:];
  mean_opt_errs = mean(optimal_errs, 1)[:];
  mean_np_errs = mean(non_private_errs, 1)[:];
  num_sub_inds = 200;
  sub_inds = round(Int64, linspace(1, n_obs, num_sub_inds));
  lower_lap_errs = zeros(num_sub_inds);
  upper_lap_errs = zeros(num_sub_inds);
  lower_opt_errs = zeros(num_sub_inds);
  upper_opt_errs = zeros(num_sub_inds);
  lower_np_errs = zeros(num_sub_inds);
  upper_np_errs = zeros(num_sub_inds);
  for ii = 1:num_sub_inds
    ind = sub_inds[ii];
    lower_lap_errs[ii] = quantile(laplace_errs[:, ind], p_value);
    upper_lap_errs[ii] = quantile(laplace_errs[:, ind], 1 - p_value);
    lower_opt_errs[ii] = quantile(optimal_errs[:, ind], p_value);
    upper_opt_errs[ii] = quantile(optimal_errs[:, ind], 1 - p_value);
    lower_np_errs[ii] = quantile(non_private_errs[:, ind], p_value);
    upper_np_errs[ii] = quantile(non_private_errs[:, ind], 1 - p_value);
  end

  figure();
  semilogy(sub_inds, mean_lap_errs[sub_inds], "b-", linewidth=2);
  semilogy(sub_inds, mean_opt_errs[sub_inds], "g-", linewidth=2);
  semilogy(sub_inds, mean_np_errs[sub_inds], "k-", linewidth=2);
  fill_between(sub_inds, lower_lap_errs, upper_lap_errs,
               color=(.8, .8, 1));
  semilogy(sub_inds, lower_lap_errs, "b-");
  semilogy(sub_inds, upper_lap_errs, "b-");
  fill_between(sub_inds, lower_opt_errs, upper_opt_errs,
               color=(.8, 1, .8));
  semilogy(sub_inds, lower_opt_errs, "g-");
  semilogy(sub_inds, upper_opt_errs, "g-");
  fill_between(sub_inds, lower_np_errs, upper_np_errs,
               color=(.8, .8, .8));
  axis([0, maximum(sub_inds),
        10^(floor(log10(minimum(mean_np_errs)))), 1]);        
  if (!isempty(filename))
    savefig(filename, bbox_inches="tight");
  end
end

