# private-vector-sampling.jl
#
# Methods for drawing samples Z, conditional on v, with the property that
#
# E[Z | v] = v
#
# and Z is differentially private for the initial vector v.

import Base.dot;

# Z = PrivatizeLtwoVector(v::Vector{Float64},
#                         true_radius::Float64,
#                         alpha::Real)
#
# Computes an alpha-differentially private realization of the vector
# v, which should satisfy norm(v) <= true_radius, with the property that
#
# E[Z | v] = v.
function PrivatizeLtwoVector(v::Vector{Float64},
                             true_radius::Real,
                             alpha::Real)
  norm_v = norm(v);
  @assert norm_v <= true_radius;
  @assert !any(isnan(v));
  dim = length(v);
  B = PrivateDataRadius(dim, true_radius, alpha);
  # Blow v up to have radius true_radius
  u = rand();
  if (norm_v > true_radius * 1e-10)
    v = true_radius * sign(.5 + norm_v / (2 * true_radius) - u) * v / norm_v;
  else
    v = randn(dim); v = true_radius * v / norm(v);
  end
  # T is 1 with probability e^alpha / (e^alpha + 1) and -1 otherwise.
  T = sign(exp(alpha) / (exp(alpha) + 1) - rand());
  rand_vec = randn(dim);
  rand_vec /= norm(rand_vec);
  while (sign(dot(rand_vec, v)) != T)
    rand_vec = randn(dim);
    rand_vec /= norm(rand_vec);
  end
  return B * rand_vec;
end

# B = PrivateDataRadius(dim::Int64, init_radius::Real, alpha::Real)
#
# Returns the radius needed for optimal L2-based sampling of
# gradients.
function PrivateDataRadius(dim::Int64, init_radius::Real, alpha::Real)
  B = (.5 * init_radius * ((e^alpha + 1) / (e^alpha - 1)) *
       dim * sqrt(pi) * exp(lgamma((dim - 1) / 2 + 1)
                            - lgamma(dim / 2 + 1)));
  return B;
end

# Z = LaplacePrivatizeLtwoVector(v::Vector{Float64}, true_radius::Real,
#                                alpha::Real)
#
# Draws an independent Laplace(2 * alpha / sqrt(dim)) vector, adds it
# to vector v, resulting in Z. Is an alpha-differentially private
# version of v as long as v has L2-norm at most true_radius.
function LaplacePrivatizeLtwoVector(v::Vector{Float64},
                                    true_radius::Real,
                                    alpha::Real)
  norm_v = norm(v);
  @assert norm_v <= true_radius;
  dim = length(v);
  W = ((2 * true_radius * sqrt(dim) / alpha) *
       sign(rand(dim) - .5) .* log(rand(dim)));
  return v + W;
end

# Z = PrivatizeLinfVector(v, true_radius, alpha)
#
# Generates a private random variable Z with expectation E[Z | v] = v,
# where Z is alpha-differentially private for v. Assumes that the
# L-infinity radius of v is bounded by true_radius.
#
# The sampling scheme is as follows: first, we flip a biased coin with
# heads probability p(alpha) = e^alpha / (e^alpha + 1). If the coin is
# heads, we uniformly sample a vector Z in {-1, 1}^d subject to Z * v
# > 0, otherwise we uniformly sample from Z * v <= 0.
#  
# The expectation of a random variable Z sampled according to Z*v > 0 is
#
#  (2/2^d) * C(d) * v
#
# where C(d) = Choose(d-1,(d-1)/2) if d is odd, C(d) = Choose(d-1,d/2)
# if d is even. This makes the expectation of the random vector Z equal to
#
#   v * (2 * p(alpha) - 1) * C(d) / 2^(d-1).
#
# Multiplying Z by an appropriate constant gives the correct expectation.
function PrivatizeLinfVector(v::Vector{Float64},
                             true_radius::Real,
                             alpha::Real)
  norm_v = norm(v, Inf);
  @assert norm_v <= true_radius;
  @assert !any(isnan(v));
  dd = length(v);
  # First, make sure v is at the boundary by independently blowing
  # coordinates up to it.
  v_new = zeros(dd);
  v_new = true_radius * sign(.5 + v / (2 * true_radius) - rand(dd));

  p_flip = exp(alpha) / (exp(alpha) + 1);
  flip = (rand() > p_flip);
  z = sign(.5 - rand(dd));
  if (flip)
    while (dot(z, v_new) > 0)
      z = sign(.5 - rand(dd));
    end
  else
    while (dot(z, v_new) < 0)
      z = sign(.5 - rand(dd));
    end
  end
  Cd = (mod(dd, 2) == 0 ?
        (2^(dd - 1) + .5 * binomial(dd, round(Int64, dd/2))) /
        binomial(dd - 1, round(Int64, dd/2)) :
        2^(dd - 1) / binomial(dd - 1, round(Int64, (dd - 1)/2)));
  z = true_radius * z * Cd / (2 * p_flip - 1);
  return z;
end

# Z = PrivatizeLinfVectorLaplace(v::Vector{Float64},
#                                true_radius::Real, alpha::Real)
#
# Uses Laplace noise addition to give a alpha-differentially private
# version of v, where E[Z | v] = v. Must be the case that true_radius
# >= norm(v, Inf).
function PrivatizeLinfVectorLaplace(v::Vector{Float64},
                                    true_radius::Real, alpha::Real)
  dd = length(v);
  @assert (norm(v, Inf) <= true_radius);
  W = (dd / alpha) * sign(rand(dd) - .5) .* log(rand(dd));
  return v + W;
end

function dot(v::SparseMatrixCSC, u::Vector)
  inds = find(v);
  dp = 0;
  for ind in inds
    dp += v[ind] * u[ind];
  end
end
